import json
import sys
from pathlib import Path

import pytest
from rdflib import Literal, URIRef
from rdflib.namespace import SDO, XSD

from vocexcel.utils import ConversionError

sys.path.append(str(Path(__file__).parent.parent.absolute() / "vocexcel"))
from vocexcel import convert

TEMPLATES_DIR_PATH = Path(__file__).parent.parent.absolute() / "vocexcel/templates"
TESTS_DATA_DIR_PATH = Path(__file__).parent.absolute() / "data"


def test_085():
    g = convert.excel_to_rdf(TESTS_DATA_DIR_PATH / "085.xlsx", output_format="graph")

    assert (
        URIRef("http://example.com/voc/myvoc"),
        SDO.status,
        URIRef("https://linked.data.gov.au/def/reg-statuses/experimental"),
    ) in g

    assert (
        URIRef("http://example.com/voc/myvoc/cat"),
        SDO.status,
        URIRef("https://linked.data.gov.au/def/reg-statuses/stable"),
    ) in g

    assert (
        URIRef("http://example.com/voc/myvoc/cat"),
        SDO.image,
        Literal("Image at L4"),
    ) in g

    assert (
        URIRef("http://example.com/voc/myvoc/dog"),
        SDO.image,
        Literal(
            "https://en.wikipedia.org/wiki/Dog#/media/File:Huskiesatrest.jpg",
            datatype=XSD.anyURI,
        ),
    ) in g


def test_085GA():
    g = convert.excel_to_rdf(TESTS_DATA_DIR_PATH / "085GA.xlsx", output_format="graph")

    assert (
        URIRef("http://example.com/voc/myvoc"),
        SDO.status,
        URIRef("https://linked.data.gov.au/def/reg-statuses/experimental"),
    ) in g

    assert (
        URIRef("http://example.com/voc/myvoc"),
        SDO.identifier,
        Literal("https://pid.geoscience.gov.au/dataset/1234", datatype=XSD.anyURI),
    ) in g


def test_085_errors(capsys):
    XL_FILE = TESTS_DATA_DIR_PATH / "085GA.xlsz"
    ERROR_TXT = "Files for conversion to RDF must be Excel files ending .xlsx"

    with pytest.raises(ValueError):
        g = convert.excel_to_rdf(XL_FILE)  # error_format = "python"

    convert.excel_to_rdf(XL_FILE, error_format="cmd")
    assert ERROR_TXT in capsys.readouterr().out

    j = convert.excel_to_rdf(XL_FILE, error_format="json")
    assert ERROR_TXT in j
    p = json.loads(j)
    assert ERROR_TXT in p["message"]

    XL_FILE = TESTS_DATA_DIR_PATH / "080.xlsx"
    ERROR_TXT = "The version of your template, 0.8.0, is not supported"

    with pytest.raises(ConversionError):
        g = convert.excel_to_rdf(XL_FILE)  # error_format = "python"

    convert.excel_to_rdf(XL_FILE, error_format="cmd")
    assert ERROR_TXT in capsys.readouterr().out

    j = convert.excel_to_rdf(XL_FILE, error_format="json")
    assert ERROR_TXT in j
    p = json.loads(j)
    assert ERROR_TXT in p["message"]

    XL_FILE = TESTS_DATA_DIR_PATH / "085_invalid.xlsx"
    ERROR_TXT = (
        "Your vocabulary has no creator. Please add it to the Concept Scheme sheet"
    )

    with pytest.raises(ConversionError):
        g = convert.excel_to_rdf(XL_FILE)  # error_format = "python"

    convert.excel_to_rdf(XL_FILE, error_format="cmd")
    assert ERROR_TXT in capsys.readouterr().out

    j = convert.excel_to_rdf(XL_FILE, error_format="json")
    assert ERROR_TXT in j
    p = json.loads(j)
    assert ERROR_TXT in p["message"]

    XL_FILE = TESTS_DATA_DIR_PATH / "085_invalid2.xlsx"
    ERROR_TXT = "Your namespace value on sheet Prefixes, cell C5 is invalid. It must start with 'http'"

    with pytest.raises(ConversionError):
        g = convert.excel_to_rdf(XL_FILE)  # error_format = "python"

    convert.excel_to_rdf(XL_FILE, error_format="cmd")
    assert ERROR_TXT in capsys.readouterr().out

    j = convert.excel_to_rdf(XL_FILE, error_format="json")
    assert ERROR_TXT in j
    p = json.loads(j)
    assert ERROR_TXT in p["message"]



import json
from pathlib import Path

import pytest
from openpyxl.workbook import Workbook
from rdflib import SDO, SKOS, Literal, URIRef

from vocexcel.convert import excel_to_rdf, rdf_to_excel
from vocexcel.utils import STATUSES, ShaclValidationError, load_workbook


def test_basic_085():
    RDF_FILE = Path(__file__).parent / "data" / "085_rdf.ttl"
    XL_FILE = RDF_FILE.with_suffix(".xlsx")
    rdf_to_excel(RDF_FILE, output_file=XL_FILE)

    g = excel_to_rdf(load_workbook(XL_FILE), output_format="graph")
    # g.serialize(destination=RDF_FILE.with_suffix(".2.ttl"), format="longturtle")

    assert (
        URIRef("https://linked.data.gov.au/def/induration-style"),
        SDO.status,
        URIRef(STATUSES["Experimental"]),
    ) in g

    assert (
        URIRef("https://linked.data.gov.au/def/induration-style"),
        SKOS.historyNote,
        Literal("Created from GA's exiting vocab for VocExcel testing"),
    ) in g

    assert (
        URIRef("https://linked.data.gov.au/def/induration-style/duricrust"),
        SKOS.relatedMatch,
        URIRef("http://example.com/fake"),
    ) in g

    assert (
        URIRef("https://linked.data.gov.au/def/induration-style/bauxitic_nodules"),
        SKOS.notation,
        Literal(
            "NO10",
            datatype=URIRef("https://linked.data.gov.au/def/induration-style/gaId"),
        ),
    ) in g

    XL_FILE.unlink(missing_ok=True)


def test_basic_085GA():
    RDF_FILE = Path(__file__).parent / "data" / "085GA_rdf.ttl"
    XL_FILE = RDF_FILE.with_suffix(".xlsx")
    rdf_to_excel(RDF_FILE, output_file=XL_FILE)

    g = excel_to_rdf(load_workbook(XL_FILE), output_format="graph")
    # g.serialize(destination=RDF_FILE.with_suffix(".2.ttl"), format="longturtle")
    assert (
        URIRef("https://linked.data.gov.au/def/induration-style"),
        SDO.status,
        URIRef(STATUSES["Experimental"]),
    ) in g

    assert (
        URIRef("https://linked.data.gov.au/def/induration-style"),
        SKOS.historyNote,
        Literal("Created from GA's exiting vocab for VocExcel testing"),
    ) in g

    XL_FILE.unlink(missing_ok=True)


def test_error_responses(capsys):
    RDF_FILE = Path(__file__).parent / "data" / "085_rdf_invalid.ttl"
    XL_FILE = RDF_FILE.with_suffix(".xlsx")

    # :completely_cemented_duricrust - no definition
    # :cs - no creator
    with pytest.raises(ShaclValidationError):
        rdf_to_excel(RDF_FILE, output_file=XL_FILE)
    Path(XL_FILE).unlink(missing_ok=True)

    with pytest.raises(ValueError):
        rdf_to_excel(RDF_FILE, output_file=XL_FILE.with_suffix(".xlsz"))  # 'python'

    rdf_to_excel(RDF_FILE, output_file=XL_FILE.with_suffix(".xlsz"), error_format="cmd")
    assert (
        "If specifying an output_file, it must end with .xlsx"
        in capsys.readouterr().out
    )

    j = rdf_to_excel(
        RDF_FILE, output_file=XL_FILE.with_suffix(".xlsz"), error_format="json"
    )
    assert "If specifying an output_file, it must end with .xlsx" in j
    p = json.loads(j)
    assert "If specifying an output_file, it must end with .xlsx" in p["message"]

    Path(XL_FILE).unlink(missing_ok=True)

    with pytest.raises(ShaclValidationError):
        rdf_to_excel(RDF_FILE)  # 'python'

    rdf_to_excel(RDF_FILE, error_format="cmd")
    assert "RDF Validation Error" in capsys.readouterr().out

    j = rdf_to_excel(RDF_FILE, error_format="json")
    assert "RDF Validation Error" in j
    p = json.loads(j)
    assert "RDF Validation Error" in p["error"]

    Path(XL_FILE).unlink(missing_ok=True)


def test_return_formats():
    RDF_FILE = Path(__file__).parent / "data" / "085_rdf.ttl"
    XL_FILE = RDF_FILE.with_suffix(".xlsx")

    # with pytest.raises(ValueError):
    #     rdf_to_excel(RDF_FILE, output_file=XL_FILE, output_format="graph")

    b = rdf_to_excel(RDF_FILE, output_file=XL_FILE, output_format="blob")
    assert isinstance(b, Workbook)

    XL_FILE.unlink(missing_ok=True)
    rdf_to_excel(RDF_FILE, output_file=XL_FILE, output_format="file")
    assert XL_FILE.exists()

    XL_FILE.unlink(missing_ok=True)



import sys
from pathlib import Path

import pytest
from rdflib import URIRef
from rdflib.namespace import RDF, SKOS

sys.path.append(str(Path(__file__).parent.parent.absolute() / "vocexcel"))
from vocexcel import convert
from vocexcel.convert_071 import (
    extract_concept_scheme,
    extract_concepts,
    extract_prefixes,
)
from vocexcel.utils import ConversionError, load_workbook

TEMPLATES_DIR_PATH = Path(__file__).parent.parent.absolute() / "vocexcel/templates"
TESTS_DATA_DIR_PATH = Path(__file__).parent.absolute() / "data"


def test_empty_template():
    with pytest.raises(ConversionError, match=".*Your vocabulary has no IRI.*"):
        convert.excel_to_rdf(TEMPLATES_DIR_PATH / "VocExcel-template-071.xlsx")


def test_empty_template_ga():
    with pytest.raises(ConversionError, match=".*Your vocabulary has no IRI.*"):
        convert.excel_to_rdf(TEMPLATES_DIR_PATH / "VocExcel-template-071-GA.xlsx")


@pytest.fixture
def get_excel():
    return load_workbook(TESTS_DATA_DIR_PATH / "071_demo.xlsx")


def test_extract_concepts(get_excel):
    wb = get_excel
    prefixes = extract_prefixes(wb["Prefixes"])
    cs, cs_iri = extract_concept_scheme(wb["Concept Scheme"], prefixes)
    cons = extract_concepts(wb["Concepts"], prefixes, cs_iri)

    concepts = []
    for c in cons.subjects(RDF.type, SKOS.Concept):
        concepts.append(c)

    assert URIRef("http://example.com/chicken-breeds/thisisabad__-URI") in concepts

    assert len(concepts) == 2



from pathlib import Path

import pytest

from vocexcel import convert

TEMPLATES_DIR_PATH = Path(__file__).parent.parent.absolute() / "vocexcel/templates"
TESTS_DATA_DIR_PATH = Path(__file__).parent.absolute() / "data"


def test_user_specified_versions():
    with pytest.raises(ValueError) as e:
        g = convert.excel_to_rdf(
            TESTS_DATA_DIR_PATH / "085.xlsx",
            output_format="graph",
            allowed_template_versions=["0.8.5.GA"],
        )
        assert "You have restricted the allowed template versions" in e

    try:
        g = convert.excel_to_rdf(
            TESTS_DATA_DIR_PATH / "085.xlsx",
            output_format="graph",
            allowed_template_versions=["0.8.X"],
        )
    except ValueError as e:
        assert (
            "You have restricted the allowed template versions to unknown template versions."
            in str(e)
        )



import json
import logging
import re
from pathlib import Path
from tempfile import SpooledTemporaryFile
from typing import Dict, Tuple, Union
from typing import Literal as TypeLiteral

import pyshacl
from colorama import Fore, Style
from openpyxl import load_workbook as _load_workbook
from openpyxl.styles import Font
from openpyxl.workbook.workbook import Workbook
from openpyxl.worksheet.worksheet import Worksheet
from pyshacl.pytypes import GraphLike
from rdflib import BNode, Graph, Literal, Namespace, Node, URIRef
from rdflib.namespace import DCTERMS, PROV, RDF, SDO, SH, SKOS, XSD

from vocexcel import profiles

EXCEL_FILE_ENDINGS = ["xlsx"]
RDF_FILE_ENDINGS = {
    ".ttl": "ttl",
    ".rdf": "xml",
    ".xml": "xml",
    ".json-ld": "json-ld",
    ".json": "json-ld",
    ".nt": "nt",
    ".n3": "n3",
}
KNOWN_FILE_ENDINGS = [str(x) for x in RDF_FILE_ENDINGS.keys()] + EXCEL_FILE_ENDINGS
KNOWN_TEMPLATE_VERSIONS = [
    "0.2.1",
    "0.3.0",
    "0.4.0",
    "0.4.1",
    "0.4.2",
    "0.4.3",
    "0.4.4",
    "0.5.0",
    "0.6.0",
    "0.6.2",
    "0.6.3",
    "0.7.0",
    "0.7.1",
    "0.8.5",
    "0.8.5.GA",
    "0.8.12",
]
LATEST_TEMPLATE = KNOWN_TEMPLATE_VERSIONS[-1]

STATUSES = {
    "Accepted": "https://linked.data.gov.au/def/reg-statuses/accepted",
    "Deprecated": "https://linked.data.gov.au/def/reg-statuses/deprecated",
    "Retired": "https://linked.data.gov.au/def/reg-statuses/retired",
    "Superseded": "https://linked.data.gov.au/def/reg-statuses/superceded",
    "Unstable": "https://linked.data.gov.au/def/reg-statuses/unstable",
    "Valid": "https://linked.data.gov.au/def/reg-statuses/valid",
    "Experimental": "https://linked.data.gov.au/def/reg-statuses/experimental",
    "Stable": "https://linked.data.gov.au/def/reg-statuses/stable",
    "Addition": "https://linked.data.gov.au/def/reg-statuses/addition",
    "Original": "https://linked.data.gov.au/def/reg-statuses/original",
    "Not Accepted": "https://linked.data.gov.au/def/reg-statuses/notAccepted",
    "Invalid": "https://linked.data.gov.au/def/reg-statuses/invalid",
    "Reserved": "https://linked.data.gov.au/def/reg-statuses/reserved",
    "Submitted": "https://linked.data.gov.au/def/reg-statuses/submitted",
}
VOCDERMODS = {
    "Direct": "https://linked.data.gov.au/def/reg-statuses/direct",
    "Extension": "https://linked.data.gov.au/def/reg-statuses/extension",
    "Subsetting & Extension": "https://linked.data.gov.au/def/reg-statuses/subsetting-and-extension",
    "None": "https://linked.data.gov.au/def/reg-statuses/none",
    "Not Applicable": "https://linked.data.gov.au/def/reg-statuses/not-applicable",
    "Relabelling": "https://linked.data.gov.au/def/reg-statuses/relabelling",
    "Subsetting": "https://linked.data.gov.au/def/reg-statuses/subsetting",
}


class ConversionError(Exception):
    pass


class ShaclValidationError(ConversionError):
    def __init__(self, message, graph: Graph):
        super().__init__(message)
        self.graph = graph


def load_workbook(
    file_path: Path | Workbook,
    error_format: TypeLiteral["python", "cmd", "json"] = "python",
) -> Workbook | str | None:
    if isinstance(file_path, Workbook):
        return file_path

    if not isinstance(
        file_path, SpooledTemporaryFile
    ) and not file_path.name.lower().endswith(tuple(EXCEL_FILE_ENDINGS)):
        error = ValueError(
            "Files for conversion to RDF must be Excel files ending .xlsx"
        )
        return return_error(error, error_format)
    return _load_workbook(filename=file_path, data_only=True)


def get_template_version(
    wb: Workbook, error_format: TypeLiteral["python", "cmd", "json"] = "python"
) -> str:
    # try 0.4.0, 0.5.0 & 0.6.x locations
    def find_version(wb: Workbook):
        try:
            if "Introduction" in wb.sheetnames:
                intro_sheet = wb["Introduction"]
                if intro_sheet["E4"].value is not None:  # 0.5.0, 0.6.x, 0.7.x, 0.8.x
                    return intro_sheet["E4"].value
                if intro_sheet["J11"].value is not None:  # 0.4.0
                    return intro_sheet["J11"].value

            # try 0.2.1 & 0.3.0 locations
            pi = wb["program info"]
            if pi["B2"].value is not None:
                return pi["B2"].value
            return None
        except Exception:
            return None

    version = find_version(wb)
    if version in KNOWN_TEMPLATE_VERSIONS:
        return version
    elif version is not None:
        error = ConversionError(
            f"The version of your template, {version}, is not supported"
        )
        return return_error(error, error_format)
    # if we get here, the template version is either unknown or can't be located
    error = ConversionError(
        "The version of the Excel template you are using cannot be determined"
    )
    return return_error(error, error_format)


def split_and_tidy_to_strings(s: str):
    s = str(s) if s is not None and not isinstance(s, str) else s
    # TODO: this may not work in list of things that contain commas. Need to consider revising
    # to allow comma-seperated values where it'll split in commas but not in things enclosed in quotes.
    if s == "" or s is None:
        return []
    else:
        return [x.strip() for x in re.split(r"[,\n]\s?", s.strip()) if x != ""]


def split_and_tidy_to_iris(s: str, prefixes, cs_iri=None):
    return [
        make_iri(ss, prefixes, cs_iri) for ss in split_and_tidy_to_strings(s)
    ]


def string_is_http_iri(s: str) -> Tuple[bool, str]:
    # returns (True, None) if the string (sort of) is an IRI
    # returns (False, message) otherwise
    messages = []
    if not s.startswith("http"):
        messages.append(
            f"HTTP IRIs must start with 'http' or 'https'. Your value was '{s}'"
        )
        if ":" in s:
            messages.append(
                f"It looks like your IRI might contain a prefix, {s.split(':')[0] + ':'}, that could not be expanded. "
                "Check it's present in the Prefixes sheet of your workbook"
            )

    if len(messages) > 0:
        return False, " and ".join(messages)
    else:
        return True, ""


def all_strings_in_list_are_iris(l_: []) -> Tuple[bool, str]:
    messages = []
    if l_ is not None:
        for item in l_:
            r = string_is_http_iri(item)
            if not r[0]:
                messages.append(f"Item {item} failed with messages {r[1]}")

    if len(messages) > 0:
        return False, " and ".join(messages)
    else:
        return True, ""


def expand_namespaces(s: str, prefixes: dict[str, Namespace]) -> Union[URIRef, str]:
    for pre in prefixes.keys():
        if s.startswith(pre):
            return URIRef(s.replace(pre, prefixes[pre]))
    if s.startswith("http"):
        return URIRef(s)
    else:
        return s


def bind_namespaces(g: Graph, prefixes: dict[str, Namespace]):
    for pre, ns in prefixes.items():
        g.bind(pre.rstrip(":"), ns)


def string_from_iri(iri):
    s = str(iri.split("/")[-1].split("#")[-1])
    s = re.sub(r"(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])", " ", s)
    s = s.title()
    s = s.replace("-", " ")

    return s


def id_from_iri(iri):
    id = str(iri.split("/")[-1].split("#")[-1])
    return Literal(id, datatype=XSD.token)


def make_agent(agent_value, agent_role, prefixes, iri_of_subject) -> Graph:
    ag = Graph()
    iri = expand_namespaces(agent_value, prefixes)
    creator_iri_conv = string_is_http_iri(str(iri))
    if not creator_iri_conv[0]:
        iri = BNode()

    if "custodian" in str(agent_role) or "orcid" in iri:
        agent_type = SDO.Person
        url_email = SDO.email
    else:
        agent_type = SDO.Organization
        url_email = SDO.url

    ag.add((iri, RDF.type, agent_type))
    ag.add((iri, url_email, Literal("", datatype=XSD.anyURI)))
    ag.add((iri, SDO.name, Literal(string_from_iri(agent_value))))
    if agent_role in [
        DCTERMS.creator,
        DCTERMS.publisher,
        DCTERMS.rightsHolder,
        SDO.creator,
        SDO.publisher,
        SDO.copyrightHolder,
    ]:
        ag.add((iri_of_subject, agent_role, iri))
    else:
        qa = BNode()
        ag.add((iri_of_subject, PROV.qualifiedAttribution, qa))
        ag.add((qa, PROV.agent, iri))
        ag.add((qa, PROV.hadRole, agent_role))

    return ag


def make_iri(s: str, prefixes: dict[str, Namespace], namespace=None):
    iri = make_clean_str(s)
    iri = expand_namespaces(iri, prefixes)
    iri = (
        namespace + iri if not iri.startswith("http") and namespace is not None else iri
    )
    is_an_iri = string_is_http_iri(str(iri))
    if not is_an_iri[0]:
        raise ConversionError(is_an_iri[1])
    return iri


def add_top_concepts(g: Graph) -> Graph:
    """For every Concept that no other Concept indicat as being narrower than it, indicate that it is a top concept"""
    cs = None
    for x in g.subjects(RDF.type, SKOS.ConceptScheme):
        cs = x
    if cs is None:
        raise ValueError("The input graph declares no SKOS ConceptScheme")

    for c in g.subjects(RDF.type, SKOS.Concept):
        if not g.value(None, SKOS.narrower, c):
            g.add((c, SKOS.topConceptOf, cs))
            g.add((cs, SKOS.hasTopConcept, c))

    return g


def validate_with_profile(
    data_graph: Union[GraphLike, str, bytes],
    profile="vocpub-46",
    error_level=1,
    message_level=1,
    log_file=None,
):
    if profile not in profiles.PROFILES.keys():
        raise ValueError(
            f"The profile chosen for conversion must be one of '{', '.join(profiles.PROFILES.keys())}' "
            f"but you selected {profile}"
        )
    allow_warnings = True if error_level > 1 else False

    # validate the RDF file
    conforms, results_graph, results_text = pyshacl.validate(
        data_graph,
        shacl_graph=str(Path(__file__).parent / f"{profile}.ttl"),
        allow_warnings=allow_warnings,
    )

    logging_level = logging.INFO

    if message_level == 3:
        logging_level = logging.ERROR
    elif message_level == 2:
        logging_level = logging.WARNING

    if log_file:
        logging.basicConfig(
            level=logging_level, format="%(message)s", filename=log_file, force=True
        )
    else:
        logging.basicConfig(level=logging_level, format="%(message)s")

    info_list = []
    warning_list = []
    violation_list = []

    from rdflib.namespace import RDF, SH

    for report in results_graph.subjects(RDF.type, SH.ValidationReport):
        for result in results_graph.objects(report, SH.result):
            result_dict = {}
            for p, o in results_graph.predicate_objects(result):
                if p == SH.focusNode:
                    result_dict["focusNode"] = str(o)
                elif p == SH.resultMessage:
                    result_dict["resultMessage"] = str(o)
                elif p == SH.resultSeverity:
                    result_dict["resultSeverity"] = str(o)
                elif p == SH.sourceConstraintComponent:
                    result_dict["sourceConstraintComponent"] = str(o)
                elif p == SH.sourceShape:
                    result_dict["sourceShape"] = str(o)
                elif p == SH.value:
                    result_dict["value"] = str(o)
            result_message_formatted = log_msg(result_dict, log_file)
            result_message = log_msg(result_dict, "placeholder")
            if result_dict["resultSeverity"] == str(SH.Info):
                logging.info(result_message_formatted)
                info_list.append(result_message)
            elif result_dict["resultSeverity"] == str(SH.Warning):
                logging.warning(result_message_formatted)
                warning_list.append(result_message)
            elif result_dict["resultSeverity"] == str(SH.Violation):
                logging.error(result_message_formatted)
                violation_list.append(result_message)

    if error_level == 3:
        error_messages = violation_list
    elif error_level == 2:
        error_messages = warning_list + violation_list
    else:  # error_level == 1
        error_messages = info_list + warning_list + violation_list

    if len(error_messages) > 0:
        raise ConversionError(
            f"The file you supplied is not valid according to the {profile} profile."
        )


def log_msg(result: Dict, log_file: str) -> str:
    from rdflib.namespace import SH

    formatted_msg = ""
    message = f"""Validation Result in {result["sourceConstraintComponent"].split(str(SH))[1]} ({result["sourceConstraintComponent"]}):
\tSeverity: sh:{result["resultSeverity"].split(str(SH))[1]}
\tSource Shape: <{result["sourceShape"]}>
\tFocus Node: <{result["focusNode"]}>
\tValue Node: <{result.get("value", "")}>
\tMessage: {result["resultMessage"]}
"""
    if result["resultSeverity"] == str(SH.Info):
        formatted_msg = (
            f"INFO: {message}"
            if log_file
            else Fore.BLUE + "INFO: " + Style.RESET_ALL + message
        )
    elif result["resultSeverity"] == str(SH.Warning):
        formatted_msg = (
            f"WARNING: {message}"
            if log_file
            else Fore.YELLOW + "WARNING: " + Style.RESET_ALL + message
        )
    elif result["resultSeverity"] == str(SH.Violation):
        formatted_msg = (
            f"VIOLATION: {message}"
            if log_file
            else Fore.RED + "VIOLATION: " + Style.RESET_ALL + message
        )
    return formatted_msg


def make_clean_str(s: str):
    chars = "*{}[]()>+!$ %&"
    for c in chars:
        s = s.replace(c, "")

    return s


def xl_hyperlink(cell, s: str | Node):
    cell.value = str(s)
    cell.hyperlink = str(s)
    cell.style = "Hyperlink"


def fill_cell_with_list_of_curies(
    cell_id: str, ws: Worksheet, g: Graph, subj: URIRef, pred: URIRef
):
    xs = []
    for x in g.objects(subject=subj, predicate=pred):
        xs.append(x)
    ws[cell_id] = ",\n".join([g.namespace_manager.curie(z) for z in xs])
    ws[cell_id].font = Font(size=14)


def return_error(
    error, error_output_format: TypeLiteral["python", "cmd", "json"] = "python"
) -> str | None:
    if error_output_format == "python":
        raise error

    if isinstance(error, ShaclValidationError):
        error_message = "RDF Validation Error"
        error_body = format_shacl_error(error.graph, error_output_format)
    else:
        error_message = type(error).__name__
        error_body = str(error)

    if error_output_format == "json":
        return json.dumps(
            {
                "error": error_message,
                "message": error_body,
            }
        )
    elif error_output_format == "cmd":
        print(f"ERROR: {error_message}")
        print(error_body)
    return None


def format_shacl_error(
    error_graph: Graph, output_format: TypeLiteral["python", "cmd", "json"] = "python"
) -> Graph | str | list:
    if output_format == "python":
        return error_graph
    else:
        individual_results = []
        for vr in error_graph.subjects(RDF.type, SH.ValidationResult):
            fn = None
            m = None
            for p, o in error_graph.predicate_objects(vr):
                if p == SH.focusNode:
                    fn = o
                elif p == SH.resultMessage:
                    m = o
            individual_results.append((fn, m))
        if output_format == "cmd":
            s = f"No. of errors: {len(individual_results)}\n\n"
            for fn, m in individual_results:
                s += f"* {fn}:\n\t{m}\n\n"
            return s
        else:  # json
            return individual_results



from pathlib import Path
from typing import Literal as TypeLiteral
from typing import Optional

from dateutil.parser import parse as date_parser
from openpyxl.styles import Alignment, Font
from openpyxl.workbook.workbook import Workbook
from openpyxl.worksheet.worksheet import Worksheet
from pyshacl import validate as shacl_validate
from rdflib import DCTERMS, BNode, Graph, Literal, Namespace, URIRef
from rdflib.namespace import OWL, PROV, RDF, RDFS, SDO, SKOS, XSD

from vocexcel.convert_070 import (
    extract_additions_concept_properties as extract_additions_concept_properties_070,
)
from vocexcel.convert_070 import extract_collections as extract_collections_070
from vocexcel.utils import (
    RDF_FILE_ENDINGS,
    STATUSES,
    VOCDERMODS,
    ConversionError,
    ShaclValidationError,
    add_top_concepts,
    bind_namespaces,
    fill_cell_with_list_of_curies,
    load_workbook,
    make_agent,
    make_iri,
    return_error,
    split_and_tidy_to_iris,
    split_and_tidy_to_strings,
    xl_hyperlink,
)

DATAROLES = Namespace("https://linked.data.gov.au/def/data-roles/")


def extract_prefixes(
    sheet: Worksheet,
    error_format: TypeLiteral["python", "cmd", "json"] = "python",
) -> dict[str, Namespace] | str | None:
    prefixes = {}
    i = 3
    while True:
        ns = sheet[f"B{i}"].value
        if ns is None:
            break
        else:
            if not ns.startswith("http"):
                error = ConversionError(
                    f"Your namespace value on sheet Prefixes, cell C{i} is invalid. It must start with 'http'"
                )
                return return_error(error, error_format)

            pre = sheet[f"A{i}"].value
            proper_pre = str(pre).strip(":") + ":" if pre is not None else ":"
            prefixes[proper_pre] = ns

        i += 1

    return prefixes


def extract_concept_scheme(
    sheet: Worksheet,
    prefixes,
    template_version="0.8.5",
    error_format: TypeLiteral["python", "cmd", "json"] = "python",
) -> tuple[Graph, str]:
    iri_s = sheet["B3"].value
    title = sheet["B4"].value
    description = sheet["B5"].value
    created = sheet["B6"].value
    modified = sheet["B7"].value
    creator = sheet["B8"].value
    publisher = sheet["B9"].value
    custodian = sheet["B10"].value
    version = str(sheet["B11"].value).strip("'")
    history_note = sheet["B12"].value
    citation = sheet["B13"].value
    derived_from = sheet["B14"].value
    voc_der_mod = sheet["B15"].value
    themes = split_and_tidy_to_strings(sheet["B16"].value)
    status = sheet["B17"].value
    if template_version == "0.8.5.GA":
        catalogue_pid = sheet["B18"].value

    if iri_s is None:
        error = ConversionError(
            "Your vocabulary has no IRI. Please add it to the Concept Scheme sheet"
        )
        return return_error(error, error_format)
    else:
        iri = make_iri(iri_s, prefixes)

    if title is None:
        error = ConversionError(
            "Your vocabulary has no title. Please add it to the Concept Scheme sheet"
        )
        return return_error(error, error_format)

    if description is None:
        error = ConversionError(
            "Your vocabulary has no description. Please add it to the Concept Scheme sheet"
        )
        return return_error(error, error_format)

    if created is None:
        error = ConversionError(
            "Your vocabulary has no created date. Please add it to the Concept Scheme sheet"
        )
        return return_error(error, error_format)

    if modified is None:
        error = ConversionError(
            "Your vocabulary has no modified date. Please add it to the Concept Scheme sheet"
        )
        return return_error(error, error_format)

    if creator is None:
        error = ConversionError(
            "Your vocabulary has no creator. Please add it to the Concept Scheme sheet"
        )
        return return_error(error, error_format)

    if publisher is None:
        error = ConversionError(
            "Your vocabulary has no publisher. Please add it to the Concept Scheme sheet"
        )
        return return_error(error, error_format)

    if history_note is None:
        error = ConversionError(
            "Your vocabulary has no History Note statement. Please add it to the Concept Scheme sheet"
        )
        return return_error(error, error_format)

    # citation

    if derived_from is not None:
        if voc_der_mod is None:
            error = ConversionError(
                "If you supply a 'Derived From' value - IRI of another vocab - "
                "you must also supply a 'Derivation Mode' value"
            )
            return return_error(error, error_format)

        if voc_der_mod not in VOCDERMODS:
            error = ConversionError(
                f"You have supplied a vocab derivation mode for your vocab of {voc_der_mod} but it is not recognised. "
                f"If supplied, it must be one of {', '.join(VOCDERMODS.keys())}"
            )
            return return_error(error, error_format)

        derived_from = make_iri(derived_from, prefixes)

    # keywords

    if status is not None and status not in STATUSES:
        error = ConversionError(
            f"You have supplied a status for your vocab of {status} but it is not recognised. "
            f"If supplied, it must be one of {', '.join(STATUSES.keys())}"
        )
        return return_error(error, error_format)

    if template_version == "0.8.5.GA":
        if catalogue_pid is None or not str(catalogue_pid).startswith(
            "https://pid.geoscience.gov.au/"
        ):
            error = ConversionError(
                "All GA vocabularies must have an eCat ID starting https://pid.geoscience.gov.au/dataset/..., assigned in the Concept Scheme metadata"
            )
            return return_error(error, error_format)

    g = Graph(bind_namespaces="rdflib")
    g.add((iri, RDF.type, SKOS.ConceptScheme))
    g.add((iri, SKOS.prefLabel, Literal(title, lang="en")))
    g.add((iri, SKOS.definition, Literal(description, lang="en")))
    g.add((iri, SDO.dateCreated, Literal(created.date(), datatype=XSD.date)))
    g.add((iri, SDO.dateModified, Literal(modified.date(), datatype=XSD.date)))
    g += make_agent(creator, SDO.creator, prefixes, iri)
    g += make_agent(publisher, SDO.publisher, prefixes, iri)

    if custodian is not None:
        for _custodian in split_and_tidy_to_strings(custodian):
            g += make_agent(_custodian, DATAROLES.custodian, prefixes, iri)
            g.bind("DATAROLES", DATAROLES)

    if version is not None:
        g.add((iri, SDO.version, Literal(str(version))))
        g.add((iri, OWL.versionIRI, URIRef(iri + "/" + str(version))))

    g.add((iri, SKOS.historyNote, Literal(history_note)))

    if citation is not None:
        if str(citation).startswith("http"):
            val = Literal(citation, datatype=XSD.anyURI)
        else:
            val = Literal(citation)

        g.add((iri, SDO.citation, val))

    if derived_from is not None:
        qd = BNode()
        g.add((iri, PROV.qualifiedDerivation, qd))
        g.add((qd, PROV.entity, URIRef(derived_from)))
        g.add((qd, PROV.hadRole, URIRef(VOCDERMODS[voc_der_mod])))

    if themes is not None:
        for theme in themes:
            try:
                theme = make_iri(theme, prefixes)
            except ConversionError:
                theme = Literal(theme)
            g.add((iri, SDO.keywords, theme))

    if status is not None:
        g.add((iri, SDO.status, URIRef(STATUSES[status])))

    if template_version == "0.8.5.GA":
        g.add((iri, SDO.identifier, Literal(catalogue_pid, datatype=XSD.anyURI)))

    bind_namespaces(g, prefixes)
    g.bind("", Namespace(str(iri) + "/"))
    return g, iri


def extract_concepts(
    sheet: Worksheet,
    prefixes,
    cs_iri,
    error_format: TypeLiteral["python", "cmd", "json"] = "python",
):
    g = Graph(bind_namespaces="rdflib")
    i = 4
    while True:
        # get values
        iri_s = sheet[f"A{i}"].value
        pref_label = sheet[f"B{i}"].value
        definition = sheet[f"C{i}"].value
        alt_labels = sheet[f"D{i}"].value
        narrower = sheet[f"E{i}"].value
        history_note = sheet[f"F{i}"].value
        citation = sheet[f"G{i}"].value
        is_defined_by = sheet[f"H{i}"].value
        status = sheet[f"I{i}"].value
        example = sheet[f"J{i}"].value
        image_url = sheet[f"K{i}"].value
        image_embedded = sheet[f"L{i}"].value

        # check values
        if iri_s is None:
            break

        iri = make_iri(iri_s, prefixes)

        if pref_label is None:
            error = ConversionError(
                f"You must provide a Preferred Label for Concept {iri_s}"
            )
            return return_error(error, error_format)

        if definition is None:
            error = ConversionError(
                f"You must provide a Definition for Concept {iri_s}"
            )
            return return_error(error, error_format)

        if status is not None and status not in STATUSES:
            error = ConversionError(
                f"You have supplied a status for your Concept of {status} but it is not recognised. "
                f"If supplied, it must be one of {', '.join(STATUSES.keys())}"
            )
            return return_error(error, error_format)

        if image_url is not None:
            if not image_url.startswith("http"):
                error = ConversionError(
                    "If supplied, an Image URL must start with 'http'"
                )
                return return_error(error, error_format)

        # TODO: test embedded mage is not too large
        # image_embedded
        if image_embedded not in [None, "#VALUE!"]:
            error = ConversionError(
                "The Image Embedded colum, you must only insert images or leave it blank. "
                f"You have an unexpected value in Cell L{i}"
            )
            return return_error(error, error_format)

        # ignore example Concepts
        if iri_s in [
            "http://example.com/earth-science",
            "http://example.com/atmospheric-science",
            "http://example.com/geology",
        ]:
            continue

        # create Graph
        g.add((iri, RDF.type, SKOS.Concept))
        g.add((iri, SKOS.inScheme, cs_iri))

        if "@" in pref_label:
            val, lang = pref_label.strip().split("@")
            g.add((iri, SKOS.prefLabel, Literal(val, lang=lang)))
        else:
            g.add((iri, SKOS.prefLabel, Literal(pref_label.strip(), lang="en")))
        g.add((iri, SKOS.definition, Literal(definition.strip(), lang="en")))

        if alt_labels is not None:
            for al in split_and_tidy_to_strings(alt_labels):
                g.add((iri, SKOS.altLabel, Literal(al, lang="en")))

        if narrower is not None:
            for n in split_and_tidy_to_iris(narrower, prefixes):
                g.add((iri, SKOS.narrower, n))
                g.add((n, SKOS.broader, iri))

        if history_note is not None:
            g.add((iri, SKOS.historyNote, Literal(history_note.strip())))

        if citation is not None:
            for _citation in split_and_tidy_to_strings(citation):
                g.add(
                    (iri, SDO.citation, Literal(_citation.strip(), datatype=XSD.anyURI))
                )

        if is_defined_by is not None:
            g.add((iri, RDFS.isDefinedBy, URIRef(is_defined_by.strip())))
        else:
            g.add((iri, RDFS.isDefinedBy, cs_iri))

        if status is not None:
            g.add((iri, SDO.status, URIRef(STATUSES[status])))

        if example is not None:
            g.add((iri, SKOS.example, Literal(str(example).strip())))

        if image_url is not None:
            g.add(
                (iri, SDO.image, Literal(str(image_url).strip(), datatype=XSD.anyURI))
            )

        if image_embedded == "#VALUE!":
            g.add((iri, SDO.image, Literal(f"Image at L{i}")))

        i += 1

    bind_namespaces(g, prefixes)
    return g


def extract_collections(
    sheet: Worksheet,
    prefixes,
    cs_iri,
    error_format: TypeLiteral["python", "cmd", "json"] = "python",
):
    return extract_collections_070(sheet, prefixes, cs_iri)


def extract_additions_concept_properties(
    sheet: Worksheet,
    prefixes,
    error_format: TypeLiteral["python", "cmd", "json"] = "python",
):
    return extract_additions_concept_properties_070(sheet, prefixes)


def excel_to_rdf(
    wb: Workbook,
    output_file: Optional[Path] = None,
    template_version: str = "0.8.5",
    output_format: TypeLiteral[
        "graph",
        "rdf",
    ] = "rdf",
    error_format: TypeLiteral["python", "cmd", "json"] = "python",
):
    """Converts an Excel workbook file to RDF

    Parameters:
        excel_file: The Excel workbook file to convert
        output_file: Optional. If set, the RDF output will be written to this file in the turtle format.
        output_format: Optional, default rdf. rdf: return serialized RDF in the turtle format, graph: return an RDFLib Graph object
        error_format: Optional, default python. the kind of errors to return: python is Python, cmd is command line-formatted string, json is stringified JSON

    Returns:
        output_format or an error in one of the error_formats
    """

    if not isinstance(wb, Workbook):
        error = ValueError(
            "Files for conversion to Excel must end with one of the RDF file formats: '{}'".format(
                "', '".join(RDF_FILE_ENDINGS.keys())
            )
        )
        return return_error(error, error_format)

    if output_file is not None:
        if not Path(output_file).suffix in RDF_FILE_ENDINGS:
            error = ValueError(
                f"If specifying output_file, it must have a file suffix in {', '.join(RDF_FILE_ENDINGS)}, not {output_file}."
            )
            return return_error(error, error_format)

    if template_version not in ["0.8.5", "0.8.5.GA"]:
        error = ValueError(
            f"This converter can only handle templates with versions 0.8.x or 0.8.x.GA, not {template_version}"
        )
        return return_error(error, error_format)

    allowed_output_formats = ["graph", "rdf"]
    if output_format not in allowed_output_formats:
        error = ValueError(
            f"If specifying output_format, it be in {', '.join(allowed_output_formats)}, not {output_format}."
        )
        return return_error(error, error_format)

    allowed_error_formats = ["python", "cmd", "json"]
    if error_format not in allowed_error_formats:
        error = ValueError(
            f"error_format must be on of {', '.join(allowed_error_formats)}, not {error_format}"
        )
        return return_error(error, error_format)

    prefixes = extract_prefixes(wb["Prefixes"], error_format)
    if not isinstance(prefixes, dict):
        return prefixes

    x = extract_concept_scheme(
        wb["Concept Scheme"], prefixes, template_version, error_format
    )
    if isinstance(x, tuple):
        cs, cs_iri = x
    else:
        return x

    cons = extract_concepts(wb["Concepts"], prefixes, cs_iri)
    if not isinstance(cons, Graph):
        return cons

    cols = extract_collections(wb["Collections"], prefixes, cs_iri)
    if not isinstance(cols, Graph):
        return cols

    extra = extract_additions_concept_properties(
        wb["Additional Concept Properties"], prefixes
    )
    if not isinstance(extra, Graph):
        return extra

    g = cs + cons + cols + extra
    g = add_top_concepts(g)
    g.bind("cs", cs_iri)

    # validate the RDF file
    shacl_graph = Graph().parse(Path(__file__).parent / "vocpub-5.1.ttl")
    v = shacl_validate(g, shacl_graph=shacl_graph, allow_warnings=True)
    if not v[0]:
        return return_error(ShaclValidationError(v[2], v[1]), error_format)

    if output_file is not None:
        g.serialize(destination=str(output_file), format="longturtle")
    else:  # print to std out
        if output_format == "graph":
            return g
        else:
            return g.serialize(format="longturtle")


def rdf_to_excel(
    rdf_file: Path,
    output_file: Optional[Path] = None,
    template_version="0.8.5",
    output_format: TypeLiteral["blob", "file"] = "file",
    error_format: TypeLiteral["python", "cmd", "json"] = "python",
):
    """Converts RDF files to Excel workbooks.

    Parameters:
        rdf_file: Required. An RDF file in one of the common formats understood by RDFLib
        output_file: Optional, default none. A name for an Excel file to output. Must end in .xlsx. Not used if output_format set to blob
        template_version: Optional, default 0.8.5. Currently only 0.8.5 and 0.8.5.GA are supported
        output_format: Optional, default file. Whether to return a binary blob (openpyxl Workbook instance) or write results to file.
        error_format: Optional, default python. the kind of errors to return: python is Python, cmd is command line-formatted string, json is stringified JSON

    Returns:
        output_format or an error in one of the error_formats
    """

    # value checkers
    if not rdf_file.name.endswith(tuple(RDF_FILE_ENDINGS.keys())):
        error = ValueError(
            "Files for conversion to Excel must end with one of the RDF file formats: '{}'".format(
                "', '".join(RDF_FILE_ENDINGS.keys())
            )
        )
        return return_error(error, error_format)

    if output_file is not None:
        if not Path(output_file).suffix == ".xlsx":
            error = ValueError("If specifying an output_file, it must end with .xlsx")
            return return_error(error, error_format)

    if template_version not in ["0.8.5", "0.8.5.GA"]:
        error = ValueError(
            f"This converter can only handle templates with versions 0.8.5 or 0.8.5.GA, not {template_version}"
        )
        return return_error(error, error_format)

    allowed_output_formats = ["blob", "file"]
    if output_format not in allowed_output_formats:
        error = ValueError(
            f"output_format must be on of {', '.join(allowed_output_formats)}, not {output_format}"
        )
        return return_error(error, error_format)

    allowed_error_formats = ["python", "cmd", "json"]
    if error_format not in allowed_error_formats:
        error = ValueError(
            f"error_format must be on of {', '.join(allowed_error_formats)}, not {error_format}"
        )
        return return_error(error, error_format)

    # load the RDF file
    try:
        g = Graph().parse(str(rdf_file), format=RDF_FILE_ENDINGS[rdf_file.suffix])
    except Exception as e:
        return return_error(e, error_format)

    g.bind("ex", "http://example.com/")
    ns = g.namespace_manager

    # validate the RDF file
    shacl_graph = Graph().parse(Path(__file__).parent / "vocpub-5.1.ttl")
    v = shacl_validate(g, shacl_graph=shacl_graph, allow_warnings=True)
    if not v[0]:
        return return_error(ShaclValidationError(v[2], v[1]), error_format)

    # load the template
    fn = (
        "VocExcel-template-085-GA.xlsx"
        if template_version == "0.8.5.GA"
        else "VocExcel-template-085.xlsx"
    )
    wb = load_workbook(Path(__file__).parent / "templates" / fn)

    # Concept Scheme
    ws = wb["Concept Scheme"]
    cs_iri = g.value(predicate=RDF.type, object=SKOS.ConceptScheme)
    ws["B3"] = cs_iri
    ws["B4"] = g.value(subject=cs_iri, predicate=SKOS.prefLabel)
    ws["B5"] = g.value(subject=cs_iri, predicate=SKOS.definition)
    ws["B5"].alignment = Alignment(wrap_text=True)
    ws["B6"] = date_parser(
        str(
            g.value(subject=cs_iri, predicate=SDO.dateCreated)
            or g.value(subject=cs_iri, predicate=DCTERMS.created)
        )
    )
    ws["B7"] = date_parser(
        str(
            g.value(subject=cs_iri, predicate=SDO.dateModified)
            or g.value(subject=cs_iri, predicate=DCTERMS.modified)
        )
    )
    xl_hyperlink(
        ws["B8"],
        g.value(subject=cs_iri, predicate=SDO.creator)
        or g.value(subject=cs_iri, predicate=DCTERMS.creator),
    )
    ws["B9"] = g.value(subject=cs_iri, predicate=SDO.publisher) or g.value(
        subject=cs_iri, predicate=DCTERMS.publisher
    )
    # custodian
    for o in g.objects(subject=cs_iri, predicate=PROV.qualifiedAttribution):
        for p, o2 in g.predicate_objects(subject=o):
            if p == DATAROLES.custodian:
                ws["B10"] = str(o2)
    ws["B11"] = g.value(subject=cs_iri, predicate=SDO.version) or g.value(
        subject=cs_iri, predicate=OWL.versionInfo
    )
    ws["B12"] = g.value(subject=cs_iri, predicate=SKOS.historyNote)
    ws["B13"] = g.value(subject=cs_iri, predicate=SDO.citation)
    for o in g.objects(subject=cs_iri, predicate=PROV.qualifiedDerivation):
        for p, o2 in g.predicate_objects(subject=o):
            if p == PROV.entity:
                ws["B14"] = str(o2)
            if p == PROV.hadRole:
                for k, v in VOCDERMODS.items():
                    if v == str(o2):
                        ws["B15"] = k
    ws["B16"] = ", ".join(
        [str(x) for x in g.objects(subject=cs_iri, predicate=SDO.keywords)]
    )
    ws["B17"] = str(g.value(subject=cs_iri, predicate=SDO.status)).split("/")[-1]
    if template_version == "0.8.5.GA":
        ws["B18"] = str(g.value(subject=cs_iri, predicate=SDO.identifier))

    # Concepts
    ws = wb["Concepts"]
    r = 4
    cs = sorted(list(g.subjects(predicate=RDF.type, object=SKOS.Concept)))
    for c in cs:
        xl_hyperlink(ws[f"A{r}"], ns.curie(c))
        ws[f"B{r}"] = g.value(subject=c, predicate=SKOS.prefLabel)
        ws[f"B{r}"].font = Font(size=14)
        ws[f"C{r}"] = g.value(subject=c, predicate=SKOS.definition)
        ws[f"C{r}"].alignment = Alignment(wrap_text=True)
        ws[f"C{r}"].font = Font(size=14)

        alt_labels = []
        for alt_label in g.objects(subject=c, predicate=SKOS.altLabel):
            alt_labels.append(alt_label)
        ws[f"D{r}"] = ",\n".join(alt_labels)
        ws[f"D{r}"].font = Font(size=14)

        for s, o in g.subject_objects(SKOS.broader):
            g.add((o, SKOS.narrower, s))
        narrowers = []
        for narrower in g.objects(subject=c, predicate=SKOS.narrower):
            narrowers.append(narrower)
        ws[f"E{r}"] = ",\n".join([ns.curie(x) for x in narrowers])
        ws[f"E{r}"].font = Font(size=14)

        hn = g.value(subject=c, predicate=SKOS.historyNote)
        if hn is not None:
            ws[f"F{r}"] = hn
            ws[f"F{r}"].font = Font(size=14)

        cit = g.value(subject=c, predicate=SDO.citation)
        if cit is not None:
            ws[f"G{r}"] = hn
            ws[f"G{r}"].font = Font(size=14)

        is_defined_by = g.value(subject=c, predicate=RDFS.isDefinedBy)
        if is_defined_by != "" and is_defined_by != cs_iri:
            xl_hyperlink(ws[f"H{r}"], is_defined_by)

        status = g.value(subject=c, predicate=SDO.status)
        if status is not None:
            for k, v in STATUSES.items():
                if v == str(status):
                    ws[f"I{r}"] = k
                    ws[f"I{r}"].font = Font(size=14)

        eg = g.value(subject=c, predicate=SKOS.example)
        if eg is not None:
            ws[f"J{r}"] = eg
            ws[f"J{r}"].font = Font(size=14)

        img = g.value(subject=c, predicate=SDO.image)
        if img is not None:
            xl_hyperlink(ws[f"K{r}"], img)

        r += 1

    r = 4
    cs = sorted(list(g.subjects(predicate=RDF.type, object=SKOS.Concept)))
    for c in cs:
        # Concepts - Additional Properties
        if (
            c,
            SKOS.relatedMatch
            | SKOS.closeMatch
            | SKOS.exactMatch
            | SKOS.narrowMatch
            | SKOS.broadMatch
            | SKOS.notation,
            None,
        ) in g:
            ws = wb["Additional Concept Properties"]
            xl_hyperlink(ws[f"A{r}"], ns.curie(c))

            fill_cell_with_list_of_curies(f"B{r}", ws, g, c, SKOS.relatedMatch)
            fill_cell_with_list_of_curies(f"C{r}", ws, g, c, SKOS.closeMatch)
            fill_cell_with_list_of_curies(f"D{r}", ws, g, c, SKOS.exactMatch)
            fill_cell_with_list_of_curies(f"E{r}", ws, g, c, SKOS.narrowMatch)
            fill_cell_with_list_of_curies(f"F{r}", ws, g, c, SKOS.broadMatch)

            notations = []
            datatypes = []
            for i in g.objects(subject=c, predicate=SKOS.notation):
                notations.append(str(i))
                datatypes.append(i.datatype)
            ws[f"G{r}"] = ",\n".join(notations)
            ws[f"G{r}"].font = Font(size=14)
            ws[f"H{r}"] = ",\n".join(datatypes)
            ws[f"H{r}"].font = Font(size=14)

            r += 1

    # Collections
    ws = wb["Collections"]
    r = 4
    cols = sorted(list(g.subjects(predicate=RDF.type, object=SKOS.Collection)))
    for col in cols:
        xl_hyperlink(ws[f"A{r}"], ns.curie(col))
        ws[f"B{r}"] = g.value(subject=col, predicate=SKOS.prefLabel)
        ws[f"B{r}"].font = Font(size=14)
        ws[f"C{r}"] = g.value(subject=col, predicate=SKOS.definition)

        fill_cell_with_list_of_curies(f"D{r}", ws, g, col, SKOS.member)

        hn = g.value(subject=col, predicate=SKOS.historyNote)
        if hn is not None:
            ws[f"F{r}"] = hn
            ws[f"F{r}"].font = Font(size=14)

        r += 1

    # Namespaces
    ws = wb["Prefixes"]
    common_prefixes = [
        "ex",
        "brick",
        "csvw",
        "dc",
        "dcat",
        "dcmitype",
        "dcterms",
        "dcam",
        "doap",
        "foaf",
        "geo",
        "odrl",
        "org",
        "prof",
        "prov",
        "qb",
        "schema",
        "sh",
        "skos",
        "sosa",
        "ssn",
        "time",
        "vann",
        "void",
        "wgs",
        "owl",
        "rdf",
        "rdfs",
        "xsd",
        "xml",
    ]
    r = 4
    for pre, ns in g.namespaces():
        if pre not in common_prefixes:
            ws[f"A{r}"] = pre
            ws[f"A{r}"].font = Font(size=14)
            xl_hyperlink(ws[f"B{r}"], ns)
            r += 1

    if output_format == "blob":
        return wb
    else:
        # save the output
        if output_file is None:
            wb.save(str(rdf_file.with_suffix(".xlsx")))
        else:
            wb.save(str(output_file))



import sys
from pathlib import Path

import pytest
from rdflib import Graph, Literal, compare
from rdflib.namespace import SKOS

sys.path.append(str(Path(__file__).parent.parent.absolute() / "vocexcel"))
from vocexcel import convert
from vocexcel.convert_060 import (
    excel_to_rdf,
    extract_additions_concept_properties,
    extract_collections,
    extract_concept_scheme,
    extract_concepts,
    extract_prefixes,
)
from vocexcel.utils import ConversionError, expand_namespaces, load_workbook

TEMPLATES_DIR_PATH = Path(__file__).parent.parent.absolute() / "vocexcel/templates"
TESTS_DATA_DIR_PATH = Path(__file__).parent.absolute() / "data"


def test_empty_template():
    with pytest.raises(ConversionError, match=".*Your vocabulary has no IRI.*"):
        convert.excel_to_rdf(TEMPLATES_DIR_PATH / "VocExcel-template-060.xlsx")


@pytest.fixture
def get_excel():
    return load_workbook(TESTS_DATA_DIR_PATH / "060_simple.xlsx")


def test_extract_prefixes(get_excel):
    g = Graph()

    wb = get_excel

    n = extract_prefixes(wb["Prefixes"])

    g.add(
        (
            expand_namespaces("ex:thing-1", n),
            SKOS.prefLabel,
            Literal("Thing 1", lang="en"),
        )
    )
    g.add(
        (
            expand_namespaces("ch:rhode-island-red", n),
            SKOS.prefLabel,
            Literal("Rhode Island Red", lang="en"),
        )
    )

    # print(g.serialize(format="longturtle"))
    expected = """
                PREFIX ex: <http://example.com/>
                PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
                
                <http://example.com/chickens/rhode-island-red>
                    skos:prefLabel "Rhode Island Red"@en ;
                .
                
                ex:thing-1
                    skos:prefLabel "Thing 1"@en ;
                .    
                """
    g2 = Graph().parse(data=expected)
    assert compare.isomorphic(g, g2)


def test_extract_concept_scheme(get_excel):
    wb = get_excel
    prefixes = extract_prefixes(wb["Prefixes"])
    cs = extract_concept_scheme(wb["Concept Scheme"], prefixes)

    # no IRI
    with pytest.raises(ConversionError) as e:
        wb = load_workbook(TESTS_DATA_DIR_PATH / "060_simple2_i.xlsx")
        prefixes = extract_prefixes(wb["Prefixes"])
        extract_concept_scheme(wb["Concept Scheme"], prefixes)

    # IRI starts with unknown prefix
    with pytest.raises(ConversionError) as e:
        wb = load_workbook(TESTS_DATA_DIR_PATH / "060_simple3_i.xlsx")
        prefixes = extract_prefixes(wb["Prefixes"])
        extract_concept_scheme(wb["Concept Scheme"], prefixes)


@pytest.mark.xfail(reason="Failing since 0.6.2.")
def test_extract_concepts(get_excel):
    wb = get_excel
    prefixes = extract_prefixes(wb["Prefixes"])
    cs, cs_iri = extract_concept_scheme(wb["Concept Scheme"], prefixes)
    cons = extract_concepts(wb["Concepts"], prefixes, cs_iri)
    expected = '''PREFIX dcterms: <http://purl.org/dc/terms/>
PREFIX ex: <http://example.com/>
PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>

ex:bantam
    a skos:Concept ;
    dcterms:source "https://en.wikipedia.org/wiki/Bantam_(poultry)"^^xsd:anyURI ;
    skos:definition "A bantam is any small variety of fowl, usually of chicken or duck."@en ;
    skos:narrower ex:silkie ;
    skos:prefLabel "Bantam"@en ;
.

ex:rhode-island-red
    a skos:Concept ;
    dcterms:provenance "Taken from Wikipedia" ;
    dcterms:source "https://en.wikipedia.org/wiki/Rhode_Island_Red"^^xsd:anyURI ;
    skos:definition """The Rhode Island Red is an American breed of domestic chicken. It is the state bird of Rhode Island."""@en ;
    skos:prefLabel "Rhode Island Red"@en ;
.

ex:silkie
    a skos:Concept ;
    skos:altLabel
        "Chinese silk chicken"@en ,
        "Silky"@en ;
    skos:definition "The Silkie (also known as the Silky or Chinese silk chicken) is a breed of chicken named for its atypically fluffy plumage, which is said to feel like silk and satin."@en ;
    skos:prefLabel "Silkie"@en ;
.
'''
    g2 = Graph().parse(data=expected)
    assert compare.isomorphic(cons, g2)
    # print((g2 - cons).serialize(format="longturtle"))
    # print("++++++++++++++")
    # print((cons - g2).serialize(format="longturtle"))

    # print(cons.serialize(format="longturtle"))


@pytest.mark.xfail(reason="Failing since 0.6.2.")
def test_extract_collections(get_excel):
    wb = get_excel
    prefixes = extract_prefixes(wb["Prefixes"])
    cs, cs_iri = extract_concept_scheme(wb["Concept Scheme"], prefixes)
    cols = extract_collections(wb["Collections"], prefixes, cs_iri)
    # print(cols.serialize(format="longturtle"))
    expected = """
        PREFIX dcterms: <http://purl.org/dc/terms/>
        PREFIX ex: <http://example.com/>
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
        
        ex:small-chickens
            a skos:Concept ;
            dcterms:provenance "Just made up" ;
            skos:definition "Breeds of small chickens"@en ;
            skos:member
                ex:bantam ,
                ex:silkie ;
            skos:prefLabel "Small Chickens"@en ;
        .    
        """
    g2 = Graph().parse(data=expected)
    # print((g2 - cols).serialize(format="longturtle"))
    # print("++++++++++++++")
    # print((cols - g2).serialize(format="longturtle"))
    # print(g2.serialize(format="longturtle"))
    assert compare.isomorphic(cols, g2)


def test_extract_additions_concept_properties(get_excel):
    wb = get_excel
    prefixes = extract_prefixes(wb["Prefixes"])
    extra = extract_additions_concept_properties(
        wb["Additional Concept Properties"], prefixes
    )
    expected = """
        PREFIX ex: <http://example.com/>
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
        
        ex:bantam
            skos:notation "ban"^^ex:other-system, "B"^^<http://system-x.com> ;
            skos:relatedMatch <http://other-voc.com/bantam> ;
        .
        """
    g2 = Graph().parse(data=expected)
    assert compare.isomorphic(extra, g2)


@pytest.mark.xfail(reason="Failing since 0.6.2.")
def test_rdf_to_excel(get_excel):
    wb = get_excel
    g = excel_to_rdf(wb, output_format="graph")
    # print(g.serialize(format="longturtle"))
    expected = """
        PREFIX cdterms: <http://purl.org/dc/terms/>
        PREFIX ch: <http://example.com/chickens/>
        PREFIX dcat: <http://www.w3.org/ns/dcat#>
        PREFIX ex: <http://example.com/>
        PREFIX isoroles: <https://linked.data.gov.au/def/data-roles/>
        PREFIX owl: <http://www.w3.org/2002/07/owl#>
        PREFIX prov: <http://www.w3.org/ns/prov#>
        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
        PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
        PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
        
        ex:chickens
            a skos:ConceptScheme ;
            cdterms:created "2022-12-01"^^xsd:date ;
            cdterms:creator [
                    a prov:Agent ;
                    rdfs:label "Nicholas Car"
                ] ;
            cdterms:identifier "chickens"^^xsd:token ;
            cdterms:modified "2022-12-01"^^xsd:date ;
            cdterms:publisher <http://linked.data.gov.au/org/agldwg> ;
            owl:versionIRI ch:0.0.1 ;
            owl:versionInfo "0.0.1" ;
            skos:definition "A vocabulary of breeds of chicken"@en ;
            skos:prefLabel "Chickens Breeds"@en ;
            prov:qualifiedAttribution [
                    dcat:hadRole isoroles:custodian ;
                    prov:agent [
                            a prov:Agent ;
                            rdfs:label "Nicholas Car"
                        ]
                ] ;
        .
        
        ex:rhode-island-red
            a skos:Concept ;
            cdterms:provenance "Taken from Wikipedia" ;
            cdterms:source "https://en.wikipedia.org/wiki/Rhode_Island_Red"^^xsd:anyURI ;
            skos:definition "The Rhode Island Red is an American breed of domestic chicken. It is the state bird of Rhode Island."@en ;
            skos:prefLabel "Rhode Island Red"@en ;
        .
        
        ex:small-chickens
            a skos:Concept ;
            cdterms:provenance "Just made up" ;
            skos:definition "Breeds of small chickens"@en ;
            skos:member
                ex:bantam ,
                ex:silkie ;
            skos:prefLabel "Small Chickens"@en ;
        .
        
        ex:bantam
            a skos:Concept ;
            skos:notation "ban"^^ex:other-system, "B"^^<http://system-x.com> ;
            cdterms:source "https://en.wikipedia.org/wiki/Bantam_(poultry)"^^xsd:anyURI ;
            skos:definition "A bantam is any small variety of fowl, usually of chicken or duck."@en ;
            skos:narrower ex:silkie ;
            skos:prefLabel "Bantam"@en ;
            skos:relatedMatch <http://other-voc.com/bantam> ;
        .
        
        <http://linked.data.gov.au/org/agldwg>
            a prov:Agent ;
            rdfs:label "Agldwg" ;
        .
        
        ex:silkie
            a skos:Concept ;
            skos:altLabel
                "Chinese silk chicken"@en ,
                "Silky"@en ;
            skos:definition "The Silkie (also known as the Silky or Chinese silk chicken) is a breed of chicken named for its atypically fluffy plumage, which is said to feel like silk and satin."@en ;
            skos:prefLabel "Silkie"@en ;
        .    
    """
    g2 = Graph().parse(data=expected)
    assert compare.isomorphic(g, g2)



from pathlib import Path
from typing import Literal as TypeLiteral
from typing import Optional

from openpyxl.workbook.workbook import Workbook
from openpyxl.worksheet.worksheet import Worksheet
from pyshacl import validate as shacl_validate
from rdflib import Graph, Literal, Namespace, URIRef
from rdflib.namespace import RDF, RDFS, SDO, SKOS, XSD

from vocexcel.convert_085 import (
    extract_prefixes as extract_prefixes_085,
    extract_concept_scheme as extract_concept_scheme_085,
    rdf_to_excel as rdf_to_excel_085,
)
from vocexcel.utils import (
    STATUSES,
    ConversionError,
    bind_namespaces,
    make_iri,
    return_error,
    split_and_tidy_to_iris,
    split_and_tidy_to_strings,
    add_top_concepts,
    RDF_FILE_ENDINGS,
    ShaclValidationError,
)

DATAROLES = Namespace("https://linked.data.gov.au/def/data-roles/")


def extract_prefixes(
    sheet: Worksheet,
    error_format: TypeLiteral["python", "cmd", "json"] = "python",
) -> dict[str, Namespace] | str | None:
    return extract_prefixes_085(sheet, error_format)


def extract_concept_scheme(
    sheet: Worksheet,
    prefixes,
    template_version="0.8.5",
    error_format: TypeLiteral["python", "cmd", "json"] = "python",
) -> tuple[Graph, str]:
    return extract_concept_scheme_085(sheet, prefixes, template_version, error_format)


def extract_concepts(
    sheet: Worksheet,
    prefixes,
    cs_iri,
    error_format: TypeLiteral["python", "cmd", "json"] = "python",
):
    g = Graph(bind_namespaces="rdflib")
    i = 4
    while True:
        # get values
        iri_s = sheet[f"A{i}"].value
        pref_label = sheet[f"B{i}"].value
        definition = sheet[f"C{i}"].value
        alt_labels = sheet[f"D{i}"].value
        narrower = sheet[f"E{i}"].value
        history_note = sheet[f"F{i}"].value
        citation = sheet[f"G{i}"].value
        is_defined_by = sheet[f"H{i}"].value
        status = sheet[f"I{i}"].value
        example = sheet[f"J{i}"].value
        image_url = sheet[f"K{i}"].value
        image_embedded = sheet[f"L{i}"].value

        # check values
        if iri_s is None:
            break

        iri = make_iri(iri_s, prefixes, namespace=cs_iri + "/")

        if pref_label is None:
            error = ConversionError(
                f"You must provide a Preferred Label for Concept {iri_s}"
            )
            return return_error(error, error_format)

        if definition is None:
            error = ConversionError(
                f"You must provide a Definition for Concept {iri_s}"
            )
            return return_error(error, error_format)

        if status is not None and status not in STATUSES:
            error = ConversionError(
                f"You have supplied a status for your Concept of {status} but it is not recognised. "
                f"If supplied, it must be one of {', '.join(STATUSES.keys())}"
            )
            return return_error(error, error_format)

        if image_url is not None:
            if not image_url.startswith("http"):
                error = ConversionError(
                    "If supplied, an Image URL must start with 'http'"
                )
                return return_error(error, error_format)

        # TODO: test embedded mage is not too large
        # image_embedded
        if image_embedded not in [None, "#VALUE!"]:
            error = ConversionError(
                "The Image Embedded colum, you must only insert images or leave it blank. "
                f"You have an unexpected value in Cell L{i}"
            )
            return return_error(error, error_format)

        # ignore example Concepts
        if iri_s in [
            "http://example.com/earth-science",
            "http://example.com/atmospheric-science",
            "http://example.com/geology",
        ]:
            continue

        # create Graph
        g.add((iri, RDF.type, SKOS.Concept))
        g.add((iri, SKOS.inScheme, cs_iri))

        if "@" in pref_label:
            val, lang = pref_label.strip().split("@")
            g.add((iri, SKOS.prefLabel, Literal(val, lang=lang)))
        else:
            g.add((iri, SKOS.prefLabel, Literal(pref_label.strip(), lang="en")))
        g.add((iri, SKOS.definition, Literal(definition.strip(), lang="en")))

        if alt_labels is not None:
            for al in split_and_tidy_to_strings(alt_labels):
                g.add((iri, SKOS.altLabel, Literal(al, lang="en")))

        if narrower is not None:
            for n in split_and_tidy_to_iris(narrower, prefixes, cs_iri + "/"):
                g.add((iri, SKOS.narrower, n))
                g.add((n, SKOS.broader, iri))

        if history_note is not None:
            g.add((iri, SKOS.historyNote, Literal(history_note.strip())))

        if citation is not None:
            for _citation in split_and_tidy_to_strings(citation):
                g.add(
                    (iri, SDO.citation, Literal(_citation.strip(), datatype=XSD.anyURI))
                )

        if is_defined_by is not None:
            g.add((iri, RDFS.isDefinedBy, URIRef(is_defined_by.strip())))
        else:
            g.add((iri, RDFS.isDefinedBy, cs_iri))

        if status is not None:
            g.add((iri, SDO.status, URIRef(STATUSES[status])))

        if example is not None:
            g.add((iri, SKOS.example, Literal(str(example).strip())))

        if image_url is not None:
            g.add(
                (iri, SDO.image, Literal(str(image_url).strip(), datatype=XSD.anyURI))
            )

        if image_embedded == "#VALUE!":
            g.add((iri, SDO.image, Literal(f"Image at L{i}")))

        i += 1

    bind_namespaces(g, prefixes)
    return g


def extract_collections(
    sheet: Worksheet,
    prefixes,
    cs_iri,
    error_format: TypeLiteral["python", "cmd", "json"] = "python",
):
    g = Graph(bind_namespaces="rdflib")
    i = 4
    while True:
        # get values
        iri_s = sheet[f"A{i}"].value
        pref_label = sheet[f"B{i}"].value
        definition = sheet[f"C{i}"].value
        members = sheet[f"D{i}"].value
        history_note = sheet[f"E{i}"].value

        # check values
        if iri_s is None:
            break

        iri = make_iri(iri_s, prefixes, namespace=cs_iri + "/")

        if pref_label is None:
            raise ConversionError(
                f"You must provide a Preferred Label for Collection {iri_s}"
            )

        if definition is None:
            raise ConversionError(
                f"You must provide a Definition for Collection {iri_s}"
            )

        # create Graph
        g.add((iri, RDF.type, SKOS.Collection))
        g.add((iri, SKOS.inScheme, cs_iri))
        g.add((iri, RDFS.isDefinedBy, cs_iri))
        if str(iri).startswith(str(cs_iri)):
            g.add((iri, RDFS.isDefinedBy, cs_iri))
        g.add((iri, SKOS.prefLabel, Literal(pref_label, lang="en")))
        g.add((iri, SKOS.definition, Literal(definition, lang="en")))

        if members is not None:
            for n in split_and_tidy_to_iris(members, prefixes, cs_iri):
                g.add((iri, SKOS.member, n))

        if history_note is not None:
            g.add((iri, SKOS.historyNote, Literal(history_note.strip())))

        i += 1

    bind_namespaces(g, prefixes)
    return g


def extract_additions_concept_properties(
    sheet: Worksheet,
    prefixes,
    cs_iri,
    error_format: TypeLiteral["python", "cmd", "json"] = "python",
):
    g = Graph(bind_namespaces="rdflib")
    i = 4
    while True:
        # get values
        iri_s = sheet[f"A{i}"].value
        related_s = sheet[f"B{i}"].value
        close_s = sheet[f"C{i}"].value
        exact_s = sheet[f"D{i}"].value
        narrow_s = sheet[f"E{i}"].value
        broad_s = sheet[f"F{i}"].value
        notation_s = sheet[f"G{i}"].value
        notation_type_s = sheet[f"H{i}"].value

        # check values
        if iri_s is None:
            break

        i += 1

        # ignore example Concepts
        if iri_s in [
            "http://example.com/geology",
        ]:
            continue

        # create Graph
        iri = make_iri(iri_s, prefixes, namespace=cs_iri + "/")

        if related_s is not None:
            related = make_iri(related_s, prefixes, namespace=cs_iri + "/")
            g.add((iri, SKOS.relatedMatch, related))

        if close_s is not None:
            close = make_iri(close_s, prefixes, namespace=cs_iri + "/")
            g.add((iri, SKOS.closeMatch, close))

        if exact_s is not None:
            exact = make_iri(exact_s, prefixes, namespace=cs_iri + "/")
            g.add((iri, SKOS.exactMatch, exact))

        if narrow_s is not None:
            narrow = make_iri(narrow_s, prefixes, namespace=cs_iri + "/")
            g.add((iri, SKOS.narrowMatch, narrow))

        if broad_s is not None:
            broad = make_iri(broad_s, prefixes, namespace=cs_iri + "/")
            g.add((iri, SKOS.broadMatch, broad))

        if notation_s is not None:
            notations = split_and_tidy_to_strings(notation_s)
            if notation_type_s is not None:
                notation_types = split_and_tidy_to_iris(notation_type_s, prefixes, cs_iri)
            else:
                notation_types = [XSD.token for x in notations]

            for j, notation in enumerate(notations):
                g.add(
                    (
                        iri,
                        SKOS.notation,
                        Literal(notation, datatype=notation_types[j]),
                    )
                )

    bind_namespaces(g, prefixes)
    return g


def excel_to_rdf(
    wb: Workbook,
    output_file: Optional[Path] = None,
    template_version: str = "0.8.5",
    output_format: TypeLiteral[
        "graph",
        "rdf",
    ] = "rdf",
    error_format: TypeLiteral["python", "cmd", "json"] = "python",
):
    """Converts an Excel workbook file to RDF

    Parameters:
        excel_file: The Excel workbook file to convert
        output_file: Optional. If set, the RDF output will be written to this file in the turtle format.
        output_format: Optional, default rdf. rdf: return serialized RDF in the turtle format, graph: return an RDFLib Graph object
        error_format: Optional, default python. the kind of errors to return: python is Python, cmd is command line-formatted string, json is stringified JSON

    Returns:
        output_format or an error in one of the error_formats
    """

    if not isinstance(wb, Workbook):
        error = ValueError(
            "Files for conversion to Excel must end with one of the RDF file formats: '{}'".format(
                "', '".join(RDF_FILE_ENDINGS.keys())
            )
        )
        return return_error(error, error_format)

    if output_file is not None:
        if not Path(output_file).suffix in RDF_FILE_ENDINGS:
            error = ValueError(
                f"If specifying output_file, it must have a file suffix in {', '.join(RDF_FILE_ENDINGS)}, not {output_file}."
            )
            return return_error(error, error_format)

    if template_version not in ["0.8.12"]:
        error = ValueError(
            f"This converter can only handle templates with versions 0.8.x or 0.8.x.GA, not {template_version}"
        )
        return return_error(error, error_format)

    allowed_output_formats = ["graph", "rdf"]
    if output_format not in allowed_output_formats:
        error = ValueError(
            f"If specifying output_format, it be in {', '.join(allowed_output_formats)}, not {output_format}."
        )
        return return_error(error, error_format)

    allowed_error_formats = ["python", "cmd", "json"]
    if error_format not in allowed_error_formats:
        error = ValueError(
            f"error_format must be on of {', '.join(allowed_error_formats)}, not {error_format}"
        )
        return return_error(error, error_format)

    prefixes = extract_prefixes(wb["Prefixes"], error_format)
    if not isinstance(prefixes, dict):
        return prefixes

    x = extract_concept_scheme(
        wb["Concept Scheme"], prefixes, template_version, error_format
    )
    if isinstance(x, tuple):
        cs, cs_iri = x
    else:
        return x

    cons = extract_concepts(wb["Concepts"], prefixes, cs_iri)
    if not isinstance(cons, Graph):
        return cons

    cols = extract_collections(wb["Collections"], prefixes, cs_iri)
    if not isinstance(cols, Graph):
        return cols

    extra = extract_additions_concept_properties(
        wb["Additional Concept Properties"], prefixes, cs_iri
    )
    if not isinstance(extra, Graph):
        return extra

    g = cs + cons + cols + extra
    g = add_top_concepts(g)
    g.bind("cs", cs_iri)

    # validate the RDF file
    shacl_graph = Graph().parse(Path(__file__).parent / "vocpub-5.1.ttl")
    v = shacl_validate(g, shacl_graph=shacl_graph, allow_warnings=True)
    if not v[0]:
        return return_error(ShaclValidationError(v[2], v[1]), error_format)

    if output_file is not None:
        g.serialize(destination=str(output_file), format="longturtle")
    else:  # print to std out
        if output_format == "graph":
            return g
        else:
            return g.serialize(format="longturtle")


def rdf_to_excel(
    rdf_file: Path,
    output_file: Optional[Path] = None,
    template_version="0.8.5",
    output_format: TypeLiteral["blob", "file"] = "file",
    error_format: TypeLiteral["python", "cmd", "json"] = "python",
):
    """Converts RDF files to Excel workbooks.

    Parameters:
        rdf_file: Required. An RDF file in one of the common formats understood by RDFLib
        output_file: Optional, default none. A name for an Excel file to output. Must end in .xlsx. Not used if output_format set to blob
        template_version: Optional, default 0.8.5. Currently only 0.8.5 and 0.8.5.GA are supported
        output_format: Optional, default file. Whether to return a binary blob (openpyxl Workbook instance) or write results to file.
        error_format: Optional, default python. the kind of errors to return: python is Python, cmd is command line-formatted string, json is stringified JSON

    Returns:
        output_format or an error in one of the error_formats
    """
    return rdf_to_excel_085(
        rdf_file, output_file, template_version, output_format, error_format
    )



import sys
from pathlib import Path

import pytest
from rdflib import Graph, Literal, URIRef, compare
from rdflib.namespace import SKOS

sys.path.append(str(Path(__file__).parent.parent.absolute() / "vocexcel"))
from vocexcel import convert
from vocexcel.utils import ConversionError

TEMPLATES_DIR_PATH = Path(__file__).parent.parent.absolute() / "vocexcel/templates"
TESTS_DATA_DIR_PATH = Path(__file__).parent.absolute() / "data"


def test_empty_template():
    with pytest.raises(
        ConversionError, match=".*7 validation errors for ConceptScheme.*"
    ):
        convert.excel_to_rdf(TEMPLATES_DIR_PATH / "VocExcel-template-040.xlsx")


@pytest.mark.xfail(
    reason="Incompatible with Pydantic v2, 40008 and 44005 are nto parsable as dates"
)
def test_simple():
    convert.excel_to_rdf(
        TESTS_DATA_DIR_PATH / "040_simple.xlsx",
        output_file=TESTS_DATA_DIR_PATH / "040_simple.ttl",
    )
    g = Graph().parse(TESTS_DATA_DIR_PATH / "040_simple.ttl")
    assert len(g) == 142
    assert (
        URIRef(
            "http://resource.geosciml.org/classifierscheme/cgi/2016.01/particletype"
        ),
        SKOS.prefLabel,
        Literal("Particle Type", lang="en"),
    ) in g, "PrefLabel for vocab is not correct"
    Path(TESTS_DATA_DIR_PATH / "040_simple.ttl").unlink()


def test_exhaustive_template_is_isomorphic():
    g1 = Graph().parse(TESTS_DATA_DIR_PATH / "040_exhaustive_comparison.ttl")
    g2 = convert.excel_to_rdf(
        TESTS_DATA_DIR_PATH / "040_exhaustive.xlsx", output_format="graph"
    )
    assert compare.isomorphic(g1, g2), "Graphs are not Isomorphic"


def test_minimal_template_is_isomorphic():
    g1 = Graph().parse(TESTS_DATA_DIR_PATH / "040_minimal_comparison.ttl")
    g2 = convert.excel_to_rdf(
        TESTS_DATA_DIR_PATH / "040_minimal.xlsx", output_format="graph"
    )
    assert compare.isomorphic(g1, g2), "Graphs are not Isomorphic"



import sys
from pathlib import Path

import vocexcel.utils

sys.path.append(str(Path(__file__).parent.parent.absolute() / "vocexcel"))
from rdflib import URIRef
from rdflib.namespace import SKOS

TESTS_DATA_DIR_PATH = Path(__file__).parent.absolute() / "data"


def test_example_complex():
    from vocexcel import models
    from vocexcel.convert_030 import (
        extract_concept_scheme,
        extract_concepts_and_collections,
    )

    wb = vocexcel.utils.load_workbook(
        Path(__file__).parent.parent.resolve()
        / "vocexcel/templates"
        / "VocExcel-template-030.xlsx"
    )

    sheet = wb["example - complex"]
    cs = extract_concept_scheme(sheet)

    concepts, collections = extract_concepts_and_collections(sheet)

    g = models.Vocabulary(
        concept_scheme=cs, concepts=concepts, collections=collections
    ).to_graph()

    top_concepts = 0
    for s, o in g.subject_objects(SKOS.topConceptOf):
        top_concepts += 1
    assert top_concepts == 3, (
        f'Processing the test vocab ("example - complex" sheet in workbook '
        f"VocExcel-template.ttl) has yielded {top_concepts} top concepts but it should have "
        f"yielded 3"
    )

    for o in g.objects(
        URIRef(
            "http://resource.geosciml.org/classifierscheme/cgi/2016.01/particletype"
        ),
        SKOS.prefLabel,
    ):
        assert str(o) == "Particle Type", (
            f'The title (preferredLabel) of the "example - complex" sheet vocab in '
            f'the VocExcel-template.ttl is {s} but should be "Particle Type"'
        )


if __name__ == "__main__":
    test_example_complex()



class Profile:
    """A Profile is a specification that constrains, extends, combines, or provides guidance or explanation about
    the usage of other specifications.

    This definition includes what are sometimes called "application profiles", "metadata application profiles", or
    "metadata profiles". In this document, "data profile" and these other variants are all referred to as just
    "profiles".

    This definition of Profiles is from The Profiles Vocabulary: https://www.w3.org/TR/dx-prof/ and objects created
    using this class are intended for use as prof:Profile instances.
    """

    def __init__(
        self,
        uri,
        label,
        comment,
        mediatypes,
        default_mediatype,
        languages=None,
        default_language="en",
        is_profile_of=None,
    ):
        """
        Constructor

        :param uri: The namespace URI for the *profile* view.
        :type uri: str
        :param label: The human-readable label for the Profile.
        :type label: str
        :param comment: A description of the Profile.
        :type comment: str
        :param mediatypes: The list of IANA MediaTypes that this Profile supports.
        :type mediatypes: list
        :param default_mediatype: The default MediaType that this Profile is delivered in.
        :type default_mediatype: str
        :param languages: A list of languages this Profile supports. Identified by ISO 639-1 2-letter language codes.
        :type languages: list
        :param default_language: The default language, by default it is 'en': English.
        :type default_language: str
        :param is_profile_of: A list of URIs (strings) that this Profile is a Profile of
        :type is_profile_of: list
        """
        self.uri = uri
        self.label = label
        self.comment = comment
        self.mediatypes = mediatypes
        self.default_mediatype = default_mediatype
        self.languages = languages if languages is not None else ["en"]
        self.default_language = default_language
        self.is_profile_of = is_profile_of

    def __str__(self):
        return self.uri


RDF_MEDIA_TYPES = [
    "text/turtle",
    "application/rdf+xml",
    "application/ld+json",
    "text/n3",
    "application/n-triples",
]


HTML_MEDIA_TYPE = "text/html"


VOC_PUB_PROFILE_46 = Profile(
    "https://w3id.org/profile/vocpub/4.6",
    "Vocabulary Publication Profile v4.6",
    "SKOS ConceptSchemes, Collections & Concepts as well as additional Dublin Core Terms & schema.org "
    "annotation properties to be used to publish vocabularies.",
    RDF_MEDIA_TYPES + [HTML_MEDIA_TYPE, "text/markdown"],
    HTML_MEDIA_TYPE,
    languages=["en"],
    default_language="en",
    is_profile_of=["https://www.w3.org/TR/skos-reference/"],
)

VOC_PUB_PROFILE_51 = Profile(
    "https://w3id.org/profile/vocpub/5.1",
    "Vocabulary Publication Profile v5.1",
    "SKOS ConceptSchemes, Collections & Concepts as well as additional Dublin Core Terms & schema.org "
    "annotation properties to be used to publish vocabularies.",
    RDF_MEDIA_TYPES + [HTML_MEDIA_TYPE, "text/markdown"],
    HTML_MEDIA_TYPE,
    languages=["en"],
    default_language="en",
    is_profile_of=["https://www.w3.org/TR/skos-reference/"],
)


PROFILES = {
    "vocpub-46": VOC_PUB_PROFILE_46,
    "vocpub-51": VOC_PUB_PROFILE_51,
}



import argparse
import sys
import warnings
from pathlib import Path
from typing import BinaryIO, Literal, Optional
from typing import Literal as TypeLiteral

from openpyxl.workbook import Workbook
from pydantic import ValidationError

warnings.simplefilter(action="ignore", category=UserWarning)

THIS_FILE_PATH = Path(__file__)
sys.path.append(str(THIS_FILE_PATH.parent.parent))

from vocexcel import models
from vocexcel.convert_021 import (
    extract_concepts_and_collections as extract_concepts_and_collections_021,
)
from vocexcel.convert_030 import (
    extract_concept_scheme as extract_concept_scheme_030,
)
from vocexcel.convert_030 import (
    extract_concepts_and_collections as extract_concepts_and_collections_030,
)
from vocexcel.convert_040 import (
    extract_concept_scheme as extract_concept_scheme_040,
)
from vocexcel.convert_040 import (
    extract_concepts_and_collections as extract_concepts_and_collections_040,
)
from vocexcel.convert_043 import (
    create_prefix_dict,
)
from vocexcel.convert_043 import (
    extract_concept_scheme as extract_concept_scheme_043,
)
from vocexcel.convert_043 import (
    extract_concepts_and_collections as extract_concepts_and_collections_043,
)
from vocexcel.convert_060 import excel_to_rdf as excel_to_rdf_060
from vocexcel.convert_063 import excel_to_rdf as excel_to_rdf_063
from vocexcel.convert_070 import excel_to_rdf as excel_to_rdf_070
from vocexcel.convert_085 import excel_to_rdf as excel_to_rdf_085
from vocexcel.convert_0812 import excel_to_rdf as excel_to_rdf_0812
from vocexcel.convert_085 import rdf_to_excel as rdf_to_excel_085
from vocexcel.utils import (
    EXCEL_FILE_ENDINGS,
    KNOWN_FILE_ENDINGS,
    KNOWN_TEMPLATE_VERSIONS,
    RDF_FILE_ENDINGS,
    ConversionError,
    get_template_version,
    load_workbook,
    return_error,
)

TEMPLATE_VERSION = None


def excel_to_rdf(
    excel_file: Path | BinaryIO,
    output_file: Optional[Path] = None,
    allowed_template_versions=None,
    output_format: Literal["rdf", "graph"] = "rdf",
    error_format: TypeLiteral["python", "cmd", "json"] = "python",
):
    """Converts an Excel workbook file to RDF

    Parameters:
        excel_file: The Excel workbook file to convert
        output_file: Optional. If set, the RDF output will be written to this file in the turtle format
        allowed_template_versions: Optional. User-specified list of allowed template versions. If not given, all versions allowed
        output_format: Optional, default rdf. rdf: return serialized RDF in the turtle format, graph: return an RDFLib Graph object
        error_format: Optional, default python. the kind of errors to return: python is Python, cmd is command line-formatted string, json is stringified JSON

    Returns:
        output_format or an error in one of the error_formats
    """
    wb = load_workbook(excel_file, error_format)
    if not isinstance(wb, Workbook):
        return wb

    actual_template_version = get_template_version(wb, error_format)

    if allowed_template_versions is not None:
        if not set(allowed_template_versions).issubset(KNOWN_TEMPLATE_VERSIONS):
            error = ValueError(
                f"You have restricted the allowed template versions to unknown template versions. Known template versions are {', '.join(KNOWN_TEMPLATE_VERSIONS)}"
            )
            return return_error(error, error_format)
        elif actual_template_version not in allowed_template_versions:
            error = ValueError(
                f"You have restricted the allowed template versions to {', '.join(allowed_template_versions)} but supplied a template of version {actual_template_version}"
            )
            return return_error(error, error_format)
    elif actual_template_version not in KNOWN_TEMPLATE_VERSIONS:
        error = ConversionError(
            f"Unknown template version: {actual_template_version}. Must be one of {', '.join(KNOWN_TEMPLATE_VERSIONS)}"
        )
        return return_error(error, error_format)

    if actual_template_version in ["0.8.12"]:
        return excel_to_rdf_0812(
            wb, output_file, actual_template_version, output_format, error_format
        )

    elif actual_template_version in ["0.8.5", "0.8.5.GA"]:
        return excel_to_rdf_085(
            wb, output_file, actual_template_version, output_format, error_format
        )

    elif actual_template_version in ["0.7.1"]:
        return excel_to_rdf_070(
            wb,
            output_file,
            output_format,
            template_version=actual_template_version,
        )

    elif actual_template_version in ["0.7.0"]:
        return excel_to_rdf_070(
            wb,
            output_file,
            output_format,
            template_version=actual_template_version,
        )

    # The way the voc is made - which Excel sheets to use - is dependent on the particular template version
    elif actual_template_version in ["0.6.2", "0.6.3"]:
        return excel_to_rdf_063(
            wb,
            output_file,
            "longturtle" if output_format == "rdf" else "graph",
            template_version=actual_template_version,
        )

    elif actual_template_version in ["0.5.0", "0.6.0", "0.6.1"]:
        return excel_to_rdf_060(
            wb,
            output_file,
            "longturtle" if output_format == "rdf" else "graph",
        )

    elif actual_template_version in ["0.4.3", "0.4.4"]:
        try:
            sheet = wb["Concept Scheme"]
            concept_sheet = wb["Concepts"]
            additional_concept_sheet = wb["Additional Concept Features"]
            collection_sheet = wb["Collections"]
            prefix_sheet = wb["Prefix Sheet"]
            prefix = create_prefix_dict(prefix_sheet)

            concepts, collections = extract_concepts_and_collections_043(
                concept_sheet, additional_concept_sheet, collection_sheet, prefix
            )
            cs = extract_concept_scheme_043(sheet, prefix)
        except ValidationError as e:
            raise ConversionError(f"ConceptScheme processing error: {e}")

    elif actual_template_version == "0.3.0" or actual_template_version == "0.2.1":
        sheet = wb["vocabulary"]
        # read from the vocabulary sheet of the workbook unless given a specific sheet

        if actual_template_version == "0.2.1":
            concepts, collections = extract_concepts_and_collections_021(sheet)
        elif actual_template_version == "0.3.0":
            concepts, collections = extract_concepts_and_collections_030(sheet)

        try:
            cs = extract_concept_scheme_030(sheet)
        except ValidationError as e:
            raise ConversionError(f"ConceptScheme processing error: {e}")

    elif (
        actual_template_version == "0.4.0"
        or actual_template_version == "0.4.1"
        or actual_template_version == "0.4.2"
    ):
        try:
            sheet = wb["Concept Scheme"]
            concept_sheet = wb["Concepts"]
            additional_concept_sheet = wb["Additional Concept Features"]
            collection_sheet = wb["Collections"]

            concepts, collections = extract_concepts_and_collections_040(
                concept_sheet, additional_concept_sheet, collection_sheet
            )
            cs = extract_concept_scheme_040(sheet)
        except ValidationError as e:
            error = ConversionError(f"ConceptScheme processing error: {e}")
            return return_error(error, error_format)

    # Build the total vocab
    vocab_graph = models.Vocabulary(
        concept_scheme=cs, concepts=concepts, collections=collections
    ).to_graph()

    if output_file is not None:
        return vocab_graph.serialize(destination=str(output_file), format="longturtle")
    else:  # print to std out
        if output_format == "graph":
            return vocab_graph
        else:
            return vocab_graph.serialize(format="longturtle")


def rdf_to_excel(
    rdf_file: Path,
    output_file: Optional[Path] = None,
    template_version="0.8.5",
    output_format: TypeLiteral["blob", "file"] = "file",
    error_format: TypeLiteral["python", "cmd", "json"] = "python",
):
    """Converts RDF files to Excel workbooks.

    Parameters:
        rdf_file: Required. An RDF file in one of the common formats understood by RDFLib
        output_file: Optional, default none. A name for an Excel file to output. Must end in .xlsx. Not used if output_format set to blob
        template_version: Optional, default 0.8.5. Currently only 0.8.5 and 0.8.5.GA are supported
        output_format: Optional, default file. Whether to return a binary blob (openpyxl Workbook instance) or write results to file.
        error_format: Optional, default python. the kind of errors to return: python is Python, cmd is command line-formatted string, json is stringified JSON

    Returns:
        output_format or an error in one of the error_formats
    """
    return rdf_to_excel_085(
        rdf_file,
        output_file,
        template_version,
        output_format,
        error_format,
    )


def main(args=None):
    if args is None:  # vocexcel run via entrypoint
        args = sys.argv[1:]

    if args is None or args == []:
        return return_error(
            ValueError("You must supply the path to a file to convert"), "cmd"
        )

    parser = argparse.ArgumentParser(
        prog="vocexcel", formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        "-i",
        "--info",
        help="The version and other info of this instance of VocExcel.",
        action="store_true",
    )

    parser.add_argument(
        "input_file",
        nargs="?",  # allow 0 or 1 file name as argument
        type=Path,
        help="The Excel file to convert to a SKOS vocabulary in RDF or an RDF file to convert to an Excel file.",
    )

    parser.add_argument(
        "-o",
        "--outputfile",
        help="An optionally-provided output file path. If not provided, output from Excel-> RDF is to standard out and RDF->Excel is input file with .xlsx file ending.",
        required=False,
    )

    args = parser.parse_args(args)

    if not args:
        parser.print_help()
        parser.exit()

    if args.info:
        # not sure what to do here, just removing the errors
        from vocexcel import __version__

        print(f"VocExel version: {__version__}")
        from vocexcel.utils import KNOWN_TEMPLATE_VERSIONS

        print(
            f"Known template versions: {', '.join(sorted(KNOWN_TEMPLATE_VERSIONS, reverse=True))}"
        )
    elif args.input_file:
        if not args.input_file.suffix.lower().endswith(tuple(KNOWN_FILE_ENDINGS)):
            error = ValueError(
                "Files for conversion must either end with .xlsx (Excel) or one of the known RDF file endings, '{}'".format(
                    "', '".join(RDF_FILE_ENDINGS.keys())
                )
            )
            return return_error(error, "cmd")

        print(f"Processing file {args.input_file}")

        # input file looks like an Excel file, so convert Excel -> RDF
        if args.input_file.suffix.lower().endswith(tuple(EXCEL_FILE_ENDINGS)):
            try:
                o = excel_to_rdf(
                    args.input_file, output_file=args.outputfile, output_format="rdf"
                )
                if args.outputfile is None:
                    print(o)
            except Exception as e:
                return return_error(e, "cmd")

        # RDF file ending, so convert RDF -> Excel
        else:
            rdf_to_excel(args.input_file, args.outputfile, error_format="cmd")

            if args.outputfile is None:
                print(f"Converted result at {args.input_file.with_suffix('.xlsx')}")
            else:
                print(f"Converted result at {args.outputfile}")
    return None


if __name__ == "__main__":
    try:
        retval = main(sys.argv[1:])
        if retval is not None:
            sys.exit(retval)
    except ValueError:
        print("You must supply a path to a file to convert")



from pathlib import Path
from typing import Literal as TypeLiteral
from typing import Optional

from openpyxl.workbook.workbook import Workbook
from openpyxl.worksheet.worksheet import Worksheet
from rdflib import BNode, Graph, Literal, Namespace, URIRef
from rdflib.namespace import DCAT, DCTERMS, OWL, PROV, RDF, RDFS, SKOS, XSD

REG = Namespace("http://purl.org/linked-data/registry#")

from vocexcel.utils import (
    STATUSES,
    VOCDERMODS,
    ConversionError,
    add_top_concepts,
    bind_namespaces,
    expand_namespaces,
    id_from_iri,
    make_agent,
    make_iri,
    split_and_tidy_to_iris,
    split_and_tidy_to_strings,
    string_is_http_iri,
    validate_with_profile,
)


def extract_prefixes(sheet: Worksheet) -> dict[str, Namespace]:
    prefixes = {}
    i = 3
    while True:
        pre = sheet[f"A{i}"].value
        if pre is None:
            break
        else:
            proper_pre = str(pre) if str(pre).endswith(":") else str(pre) + ":"
            prefixes[proper_pre] = sheet[f"B{i}"].value

        i += 1

    return prefixes


def extract_concept_scheme(
    sheet: Worksheet, prefixes, template_version="0.6.3"
) -> tuple[Graph, str]:
    iri_s = sheet["B3"].value
    title = sheet["B4"].value
    description = sheet["B5"].value
    created = sheet["B6"].value
    modified = sheet["B7"].value
    creator = sheet["B8"].value
    publisher = sheet["B9"].value
    if template_version == "0.6.2":
        custodian = sheet["B12"].value
        version = str(sheet["B10"].value).strip("'")
        history_note = sheet["B11"].value
        status = None
        derived_from = None
        voc_der_mod = None
        themes = None
    else:  # 0.6.3
        custodian = sheet["B10"].value
        version = str(sheet["B11"].value).strip("'")
        history_note = sheet["B12"].value
        status = sheet["B13"].value
        derived_from = sheet["B14"].value
        voc_der_mod = sheet["B15"].value
        themes = split_and_tidy_to_strings(sheet["B16"].value)

    if iri_s is None:
        raise ConversionError(
            "Your vocabulary has no IRI. Please add it to the Concept Scheme sheet"
        )
    else:
        iri = make_iri(iri_s, prefixes)

    if title is None:
        raise ConversionError(
            "Your vocabulary has no title. Please add it to the Concept Scheme sheet"
        )

    if description is None:
        raise ConversionError(
            "Your vocabulary has no description. Please add it to the Concept Scheme sheet"
        )

    if created is None:
        raise ConversionError(
            "Your vocabulary has no created date. Please add it to the Concept Scheme sheet"
        )

    if modified is None:
        raise ConversionError(
            "Your vocabulary has no modified date. Please add it to the Concept Scheme sheet"
        )

    if creator is None:
        raise ConversionError(
            "Your vocabulary has no creator. Please add it to the Concept Scheme sheet"
        )

    if publisher is None:
        raise ConversionError(
            "Your vocabulary has no publisher. Please add it to the Concept Scheme sheet"
        )

    if history_note is None:
        raise ConversionError(
            "Your vocabulary has no History Note statement. Please add it to the Concept Scheme sheet"
        )

    if status is not None and status not in STATUSES:
        raise ConversionError(
            f"You have supplied a status for your vocab of {status} but it is not recognised. "
            f"If supplied, it must be one of {', '.join(STATUSES.keys())}"
        )

    if derived_from is not None:
        if voc_der_mod is None:
            raise ConversionError(
                "If you supply a 'Derived From' value - IRI of another vocab - "
                "you must also supply a 'Derivation Mode' value"
            )

        if voc_der_mod not in VOCDERMODS:
            raise ConversionError(
                f"You have supplied a vocab derivation mode for your vocab of {voc_der_mod} but it is not recognised. "
                f"If supplied, it must be one of {', '.join(VOCDERMODS.keys())}"
            )

        derived_from = make_iri(derived_from, prefixes)

    g = Graph(bind_namespaces="rdflib")
    g.add((iri, RDF.type, SKOS.ConceptScheme))
    g.add((iri, SKOS.prefLabel, Literal(title, lang="en")))
    g.add((iri, SKOS.definition, Literal(description, lang="en")))
    g.add((iri, DCTERMS.created, Literal(created.date(), datatype=XSD.date)))
    g.add((iri, DCTERMS.modified, Literal(modified.date(), datatype=XSD.date)))
    g.add((iri, SKOS.historyNote, Literal(history_note, lang="en")))

    g += make_agent(creator, DCTERMS.creator, prefixes, iri)

    g += make_agent(publisher, DCTERMS.publisher, prefixes, iri)

    if custodian is not None:
        for _custodian in split_and_tidy_to_strings(custodian):
            ISOROLES = Namespace("https://linked.data.gov.au/def/data-roles/")
            g += make_agent(_custodian, ISOROLES.custodian, prefixes, iri)
            g.bind("isoroles", ISOROLES)

    if version is not None:
        g.add((iri, OWL.versionInfo, Literal(str(version))))
        g.add((iri, OWL.versionIRI, URIRef(iri + "/" + str(version))))

    if status is not None:
        g.add((iri, REG.status, URIRef(STATUSES[status])))

    if derived_from is not None:
        qd = BNode()
        g.add((iri, PROV.qualifiedDerivation, qd))
        g.add((qd, PROV.entity, URIRef(derived_from)))
        g.add((qd, PROV.hadRole, URIRef(VOCDERMODS[voc_der_mod])))

    if themes is not None:
        for theme in themes:
            try:
                theme = make_iri(theme, prefixes)
            except ConversionError:
                theme = Literal(theme)
            g.add((iri, DCAT.theme, theme))

    # auto-created
    g.add((iri, DCTERMS.identifier, id_from_iri(iri)))

    bind_namespaces(g, prefixes)
    return g, iri


def extract_concepts(sheet: Worksheet, prefixes, cs_iri) -> Graph:
    g = Graph(bind_namespaces="rdflib")
    i = 4
    while True:
        # get values
        iri_s = sheet[f"A{i}"].value
        pref_label = sheet[f"B{i}"].value
        definition = sheet[f"C{i}"].value
        alt_labels = sheet[f"D{i}"].value
        narrower = sheet[f"E{i}"].value
        history_note = sheet[f"F{i}"].value
        source = sheet[f"G{i}"].value
        home = sheet[f"H{i}"].value

        # check values
        if iri_s is None:
            break

        iri = expand_namespaces(iri_s, prefixes)
        iri_conv = string_is_http_iri(str(iri))
        if not iri_conv[0]:
            raise ConversionError(iri_conv[1])

        if pref_label is None:
            raise ConversionError(
                f"You must provide a Preferred Label for Concept {iri_s}"
            )

        if definition is None:
            raise ConversionError(f"You must provide a Definition for Concept {iri_s}")

        i += 1

        # ignore example Concepts
        if iri_s in [
            "http://example.com/earth-science",
            "http://example.com/atmospheric-science",
            "http://example.com/geology",
        ]:
            continue

        # create Graph
        g.add((iri, RDF.type, SKOS.Concept))
        g.add((iri, SKOS.inScheme, cs_iri))
        if str(iri).startswith(str(cs_iri)):
            g.add((iri, RDFS.isDefinedBy, cs_iri))
        if "@" in pref_label:
            val, lang = pref_label.strip().split("@")
            g.add((iri, SKOS.prefLabel, Literal(val, lang=lang)))
        else:
            g.add((iri, SKOS.prefLabel, Literal(pref_label.strip(), lang="en")))
        g.add((iri, SKOS.definition, Literal(definition.strip(), lang="en")))

        if alt_labels is not None:
            for al in split_and_tidy_to_strings(alt_labels):
                g.add((iri, SKOS.altLabel, Literal(al, lang="en")))

        if narrower is not None:
            for n in split_and_tidy_to_iris(narrower, prefixes):
                g.add((iri, SKOS.narrower, n))

        if history_note is not None:
            g.add((iri, SKOS.historyNote, Literal(history_note.strip())))

        if source is not None:
            for _source in split_and_tidy_to_strings(source):
                g.add(
                    (iri, DCTERMS.source, Literal(_source.strip(), datatype=XSD.anyURI))
                )

        if home is not None:
            g.add((iri, RDFS.isDefinedBy, URIRef(home.strip())))

    bind_namespaces(g, prefixes)
    return g


def extract_collections(sheet: Worksheet, prefixes, cs_iri) -> Graph:
    g = Graph(bind_namespaces="rdflib")
    i = 4
    while True:
        # get values
        iri_s = sheet[f"A{i}"].value
        pref_label = sheet[f"B{i}"].value
        definition = sheet[f"C{i}"].value
        members = sheet[f"D{i}"].value
        history_note = sheet[f"E{i}"].value

        # check values
        if iri_s is None:
            break

        iri = expand_namespaces(iri_s, prefixes)
        iri_conv = string_is_http_iri(str(iri))
        if not iri_conv[0]:
            raise ConversionError(iri_conv[1])

        if pref_label is None:
            raise ConversionError(
                f"You must provide a Preferred Label for Collection {iri_s}"
            )

        if definition is None:
            raise ConversionError(
                f"You must provide a Definition for Collection {iri_s}"
            )

        # create Graph
        g.add((iri, RDF.type, SKOS.Collection))
        g.add((iri, SKOS.inScheme, cs_iri))
        g.add((iri, RDFS.isDefinedBy, cs_iri))
        if str(iri).startswith(str(cs_iri)):
            g.add((iri, RDFS.isDefinedBy, cs_iri))
        g.add((iri, SKOS.prefLabel, Literal(pref_label, lang="en")))
        g.add((iri, SKOS.definition, Literal(definition, lang="en")))

        if members is not None:
            for n in split_and_tidy_to_iris(members, prefixes):
                g.add((iri, SKOS.member, n))

        if history_note is not None:
            g.add((iri, SKOS.historyNote, Literal(history_note.strip())))

        i += 1

    bind_namespaces(g, prefixes)
    return g


def extract_additions_concept_properties(sheet: Worksheet, prefixes) -> Graph:
    g = Graph(bind_namespaces="rdflib")
    i = 4
    while True:
        # get values
        iri_s = sheet[f"A{i}"].value
        related_s = sheet[f"B{i}"].value
        close_s = sheet[f"C{i}"].value
        exact_s = sheet[f"D{i}"].value
        narrow_s = sheet[f"E{i}"].value
        broad_s = sheet[f"F{i}"].value
        notation_s = sheet[f"G{i}"].value
        notation_type_s = sheet[f"H{i}"].value

        # check values
        if iri_s is None:
            break

        i += 1

        # ignore example Concepts
        if iri_s in [
            "http://example.com/geology",
        ]:
            continue

        # create Graph
        iri = make_iri(iri_s, prefixes)
        if related_s is not None:
            related = make_iri(related_s, prefixes)
            g.add((iri, SKOS.relatedMatch, related))

        if close_s is not None:
            close = make_iri(close_s, prefixes)
            g.add((iri, SKOS.closeMatch, close))

        if exact_s is not None:
            exact = make_iri(exact_s, prefixes)
            g.add((iri, SKOS.exactMatch, exact))

        if narrow_s is not None:
            narrow = make_iri(narrow_s, prefixes)
            g.add((iri, SKOS.narrowMatch, narrow))

        if broad_s is not None:
            broad = make_iri(broad_s, prefixes)
            g.add((iri, SKOS.broadMatch, broad))

        if notation_s is not None:
            if notation_type_s is not None:
                notation_type = make_iri(notation_type_s, prefixes)
            else:
                notation_type = XSD.token
            g.add(
                (
                    iri,
                    SKOS.notation,
                    Literal(notation_s, datatype=notation_type),
                )
            )

    bind_namespaces(g, prefixes)
    return g


def excel_to_rdf(
    wb: Workbook,
    output_file_path: Optional[Path] = None,
    output_format: TypeLiteral[
        "longturtle", "turtle", "xml", "json-ld", "graph"
    ] = "longturtle",
    validate: bool = False,
    profile="vocpub-43",
    error_level=1,
    message_level=1,
    log_file: Optional[Path] = None,
    template_version="0.6.3",
):
    prefixes = extract_prefixes(wb["Prefixes"])
    cs, cs_iri = extract_concept_scheme(
        wb["Concept Scheme"], prefixes, template_version
    )
    cons = extract_concepts(wb["Concepts"], prefixes, cs_iri)
    cols = extract_collections(wb["Collections"], prefixes, cs_iri)
    extra = extract_additions_concept_properties(
        wb["Additional Concept Properties"], prefixes
    )

    g = cs + cons + cols + extra
    g = add_top_concepts(g)
    g.bind("cs", cs_iri)
    g.bind("reg", REG)

    if validate:
        validate_with_profile(
            g,
            profile=profile,
            error_level=error_level,
            message_level=message_level,
            log_file=log_file,
        )

    if output_file_path is not None:
        g.serialize(destination=str(output_file_path), format=output_format)
    else:  # print to std out
        if output_format == "graph":
            return g
        else:
            return g.serialize(format=output_format)



from pathlib import Path
from typing import Literal as TypeLiteral
from typing import Optional

from openpyxl.workbook.workbook import Workbook
from openpyxl.worksheet.worksheet import Worksheet
from rdflib import BNode, Graph, Literal, Namespace, URIRef
from rdflib.namespace import OWL, PROV, RDF, SDO, SKOS, XSD

REG = Namespace("http://purl.org/linked-data/registry#")

from vocexcel.convert_070 import excel_to_rdf as excel_to_rdf_070
from vocexcel.convert_070 import (
    extract_additions_concept_properties as extract_additions_concept_properties_070,
)
from vocexcel.convert_070 import extract_collections as extract_collections_070

# from vocexcel.convert_070 import extract_concept_scheme as extract_concept_scheme_070
from vocexcel.convert_070 import extract_concepts as extract_concepts_070
from vocexcel.convert_070 import extract_prefixes as extract_prefixes_070
from vocexcel.utils import (
    STATUSES,
    VOCDERMODS,
    ConversionError,
    bind_namespaces,
    make_agent,
    make_iri,
    split_and_tidy_to_strings,
)


def extract_prefixes(sheet: Worksheet):
    return extract_prefixes_070(sheet)


def extract_concept_scheme(
    sheet: Worksheet, prefixes, template_version="0.7.1"
) -> tuple[Graph, str]:
    iri_s = sheet["B3"].value
    title = sheet["B4"].value
    description = sheet["B5"].value
    created = sheet["B6"].value
    modified = sheet["B7"].value
    creator = sheet["B8"].value
    publisher = sheet["B9"].value
    custodian = sheet["B10"].value
    version = str(sheet["B11"].value).strip("'")
    history_note = sheet["B12"].value
    citation = sheet["B13"].value
    derived_from = sheet["B14"].value
    voc_der_mod = sheet["B15"].value
    themes = split_and_tidy_to_strings(sheet["B16"].value)
    status = sheet["B17"].value
    catalogue_pid = sheet["B18"].value

    if iri_s is None:
        raise ConversionError(
            "Your vocabulary has no IRI. Please add it to the Concept Scheme sheet"
        )
    else:
        iri = make_iri(iri_s, prefixes)

    if title is None:
        raise ConversionError(
            "Your vocabulary has no title. Please add it to the Concept Scheme sheet"
        )

    if description is None:
        raise ConversionError(
            "Your vocabulary has no description. Please add it to the Concept Scheme sheet"
        )

    if created is None:
        raise ConversionError(
            "Your vocabulary has no created date. Please add it to the Concept Scheme sheet"
        )

    if modified is None:
        raise ConversionError(
            "Your vocabulary has no modified date. Please add it to the Concept Scheme sheet"
        )

    if creator is None:
        raise ConversionError(
            "Your vocabulary has no creator. Please add it to the Concept Scheme sheet"
        )

    if publisher is None:
        raise ConversionError(
            "Your vocabulary has no publisher. Please add it to the Concept Scheme sheet"
        )

    if history_note is None:
        raise ConversionError(
            "Your vocabulary has no History Note statement. Please add it to the Concept Scheme sheet"
        )

    # citation

    if derived_from is not None:
        if voc_der_mod is None:
            raise ConversionError(
                "If you supply a 'Derived From' value - IRI of another vocab - "
                "you must also supply a 'Derivation Mode' value"
            )

        if voc_der_mod not in VOCDERMODS:
            raise ConversionError(
                f"You have supplied a vocab derivation mode for your vocab of {voc_der_mod} but it is not recognised. "
                f"If supplied, it must be one of {', '.join(VOCDERMODS.keys())}"
            )

        derived_from = make_iri(derived_from, prefixes)

    # keywords

    if status is not None and status not in STATUSES:
        raise ConversionError(
            f"You have supplied a status for your vocab of {status} but it is not recognised. "
            f"If supplied, it must be one of {', '.join(STATUSES.keys())}"
        )

    # catalogue PID

    g = Graph(bind_namespaces="rdflib")
    g.add((iri, RDF.type, SKOS.ConceptScheme))
    g.add((iri, SKOS.prefLabel, Literal(title, lang="en")))
    g.add((iri, SKOS.definition, Literal(description, lang="en")))
    g.add((iri, SDO.dateCreated, Literal(created.date(), datatype=XSD.date)))
    g.add((iri, SDO.dateModified, Literal(modified.date(), datatype=XSD.date)))
    g += make_agent(creator, SDO.creator, prefixes, iri)
    g += make_agent(publisher, SDO.publisher, prefixes, iri)

    if custodian is not None:
        for _custodian in split_and_tidy_to_strings(custodian):
            DATAROLES = Namespace("https://linked.data.gov.au/def/data-roles/")
            g += make_agent(_custodian, DATAROLES.custodian, prefixes, iri)
            g.bind("DATAROLES", DATAROLES)

    if version is not None:
        g.add((iri, SDO.version, Literal(str(version))))
        g.add((iri, OWL.versionIRI, URIRef(iri + "/" + str(version))))

    g.add((iri, SKOS.historyNote, Literal(history_note, lang="en")))

    if citation is not None:
        if str(citation).startswith("http"):
            val = Literal(citation, datatype=XSD.anyURI)
        else:
            val = Literal(citation)

        g.add((iri, SDO.citation, val))

    if derived_from is not None:
        qd = BNode()
        g.add((iri, PROV.qualifiedDerivation, qd))
        g.add((qd, PROV.entity, URIRef(derived_from)))
        g.add((qd, PROV.hadRole, URIRef(VOCDERMODS[voc_der_mod])))

    if themes is not None:
        for theme in themes:
            try:
                theme = make_iri(theme, prefixes)
            except ConversionError:
                theme = Literal(theme)
            g.add((iri, SDO.keywords, theme))

    if status is not None:
        g.add((iri, REG.status, URIRef(STATUSES[status])))

    if catalogue_pid is not None:
        if str(catalogue_pid).startswith("http"):
            val = Literal(catalogue_pid, datatype=XSD.anyURI)
        else:
            val = Literal(catalogue_pid)

        g.add((iri, SDO.identifier, val))

    bind_namespaces(g, prefixes)
    g.bind("", Namespace(str(iri) + "/"))
    return g, iri


def extract_concepts(sheet: Worksheet, prefixes, cs_iri):
    return extract_concepts_070(sheet, prefixes, cs_iri)


def extract_collections(sheet: Worksheet, prefixes, cs_iri):
    return extract_collections_070(sheet, prefixes, cs_iri)


def extract_additions_concept_properties(sheet: Worksheet, prefixes):
    return extract_additions_concept_properties_070(sheet, prefixes)


def excel_to_rdf(
    wb: Workbook,
    output_file_path: Optional[Path] = None,
    output_format: TypeLiteral[
        "longturtle", "turtle", "xml", "json-ld", "graph"
    ] = "longturtle",
    validate: bool = False,
    error_level=1,
    message_level=1,
    log_file: Optional[Path] = None,
):
    return excel_to_rdf_070(
        wb,
        output_file_path,
        output_format,
        validate,
        profile="vocpub-410",
        error_level=error_level,
        message_level=message_level,
        log_file=log_file,
        template_version="0.7.1",
    )



from typing import List, Tuple

from openpyxl.worksheet.worksheet import Worksheet
from pydantic import ValidationError

from vocexcel import models
from vocexcel.utils import ConversionError, split_and_tidy_to_strings


def extract_concepts_and_collections(
    s: Worksheet,
) -> Tuple[List[models.Concept], List[models.Collection]]:
    concepts = []
    collections = []
    process_concept = False
    process_collection = False
    for col in s.iter_cols(max_col=1):
        for cell in col:
            row = cell.row
            if cell.value == "Concept URI":
                process_concept = True
            elif cell.value == "Collection URI":
                process_concept = False
                process_collection = True
            elif process_concept:
                if cell.value is None:
                    pass
                else:
                    try:
                        c = models.Concept(
                            uri=s[f"A{row}"].value,
                            pref_label=s[f"B{row}"].value,
                            alt_labels=split_and_tidy_to_strings(s[f"C{row}"].value),
                            pl_language_code=split_and_tidy_to_strings(
                                s[f"D{row}"].value
                            ),
                            definition=s[f"E{row}"].value,
                            children=split_and_tidy_to_strings(s[f"F{row}"].value),
                            other_ids=split_and_tidy_to_strings(s[f"G{row}"].value),
                            home_vocab_uri=s[f"H{row}"].value,
                            provenance=s[f"I{row}"].value,
                        )

                        concepts.append(c)
                    except ValidationError as e:
                        raise ConversionError(
                            f"Concept processing error potentially at row {row}, with error: {e}"
                        )
            elif process_collection:
                if cell.value is None:
                    pass
                else:
                    try:
                        c = models.Collection(
                            uri=s[f"A{row}"].value,
                            pref_label=s[f"B{row}"].value,
                            definition=s[f"C{row}"].value,
                            members=split_and_tidy_to_strings(s[f"D{row}"].value),
                            provenance=s[f"E{row}"].value,
                        )
                        collections.append(c)
                    except ValidationError as e:
                        raise ConversionError(
                            f"Collection processing error, row {row}, error: {e}"
                        )
            elif cell.value is None:
                pass

    return concepts, collections



from importlib import metadata

try:
    from dotenv import load_dotenv

    load_dotenv()
except ImportError:
    ...


def get_version() -> str:
    try:
        version = metadata.version("vocexcel")
    except metadata.PackageNotFoundError:
        version = "0.0.0-dev.0"

    return version


class Settings:
    VOCEXCEL_VERSION = get_version()



import sys
from pathlib import Path

sys.path.append(str(Path(__file__).parent.parent.absolute() / "vocexcel"))
from pathlib import Path

from rdflib import Graph

from vocexcel import convert

TESTS_DATA_DIR_PATH = Path(__file__).parent.absolute() / "data"


def test_countrycodes():
    convert.excel_to_rdf(
        TESTS_DATA_DIR_PATH / "030_languages.xlsx",
        output_file=TESTS_DATA_DIR_PATH / "030_languages.ttl",
    )

    # file eg-languages-valid.ttl should have been created
    g = Graph().parse(TESTS_DATA_DIR_PATH / "030_languages.ttl")
    assert len(g) == 4940

    # clean up
    Path.unlink(TESTS_DATA_DIR_PATH / "030_languages.ttl", missing_ok=True)



import os
import sys
from pathlib import Path

from pyshacl import validate

sys.path.append(str(Path(__file__).parent.parent.absolute() / "vocexcel"))
from vocexcel.convert import excel_to_rdf

TESTS_DATA_DIR_PATH = Path(__file__).parent.absolute() / "data"


def test_validate_070_vocpub_410():
    input_exel_file_path = TESTS_DATA_DIR_PATH / "070_long.xlsx"
    validator_file_path = TESTS_DATA_DIR_PATH / "vocpub-4.10.ttl"
    output_file = TESTS_DATA_DIR_PATH / "test_validation_output.ttl"

    excel_to_rdf(input_exel_file_path, output_file=output_file)

    c, rg, rt = validate(
        str(output_file), shacl_graph=str(validator_file_path), allow_warnings=True
    )

    os.unlink(Path(output_file))

    assert c


def test_validate_071_vocpub_410():
    input_exel_file_path = TESTS_DATA_DIR_PATH / "071_demo.xlsx"
    validator_file_path = TESTS_DATA_DIR_PATH / "vocpub-4.10.ttl"
    output_file = TESTS_DATA_DIR_PATH / "test_validation_output.ttl"

    excel_to_rdf(input_exel_file_path, output_file=output_file)

    c, rg, rt = validate(
        str(output_file), shacl_graph=str(validator_file_path), allow_warnings=True
    )

    os.unlink(Path(output_file))

    assert c



from pathlib import Path
from typing import Literal, Optional

from googleapiclient.discovery import build
from pydantic import ValidationError

try:
    import models
    import profiles
    from utils import ConversionError, split_and_tidy_to_strings
except:
    import sys

    sys.path.append("..")
    from vocexcel import models, profiles
    from vocexcel.convert import log_msg, validate_with_profile
    from vocexcel.utils import (
        ConversionError,
        log_msg,
        split_and_tidy_to_strings,
        validate_with_profile,
    )

ACCEPTED_TEMPLATE_VERSIONS = ["0.4.3"]
SPREADSHEET_ID = None
RANGE = None
KEY = None
client = None
sheet = None


def get_spreadsheetid_from_URI(uri: str) -> str:
    try:
        split = uri.split("/d/")
        split2 = split[1].split("/edit")
        spreadsheetid = split2[0]
    except Exception:
        raise Exception("There seems to be something wrong with the URI supplied")
    return spreadsheetid


def only_one_colon_in_str(string: str):
    counter = 0
    for character in string:
        if character == ":":
            counter += 1
    if counter != 1:
        return False
    return True


def using_prefix_list_output(cell_value: list, prefix: dict):
    variables = []
    for i in cell_value:
        if cell_value is not None:
            try:
                c = i
                if c.startswith("http"):
                    pass
                elif only_one_colon_in_str(i):
                    split_c_prefix = c.split(":")[0]
                    split_c_added_component = c.split(":")[1]

                    if split_c_prefix in prefix:
                        c = prefix[split_c_prefix] + split_c_added_component
                    else:
                        print(
                            f"The prefix used: {split_c_prefix} isn't included in the prefix sheet"
                        )
                        raise Exception
                else:
                    print(
                        f"Something doesn't look right with this cell_value {cell_value}. "
                        f"Likely this isn't in http form or more colons are used than normal"
                    )
                    raise Exception
                variables.append(c)
            except Exception as e:
                print(e)
                exit(1)
    return variables


def using_prefix_non_list_output(cell_value, prefix: dict):
    c = cell_value
    if c is not None:
        if c.startswith("http"):
            pass
        elif only_one_colon_in_str(c):
            split_c_prefix = c.split(":")[0]
            split_c_added_component = c.split(":")[1]

            if split_c_prefix in prefix:
                c = prefix[split_c_prefix] + split_c_added_component
            else:
                print(
                    f"The prefix used: '{split_c_prefix}' isn't included in the prefix sheet"
                )
                raise Exception
        else:
            print(
                f"Something doesn't look right with the cell value: {cell_value}. "
                f"Likely this isn't in http form or more colons are used than normal"
            )
            raise Exception
    return c


def get_template_version_gsheets(spreadsheet_id: str):
    # Load workbook
    try:
        intro_sheet = (
            sheet.values()
            .get(spreadsheetId=spreadsheet_id, range="Introduction!J11")
            .execute()
        )
        template_value = intro_sheet.get("values")[0][0]
        return template_value
    except Exception:
        raise Exception(
            "The version of the Google Template you are using is not accepted/can't be determined"
        )


def read_cell(sheet_dict: dict, n: int, i: Optional[int] = 0):
    try:
        if sheet_dict.get("values")[n]:
            return sheet_dict.get("values")[n][i]
        else:
            return None
    except IndexError:
        return None


def rows_filled_out(c_sheet: dict):
    variable = len(c_sheet.get("values"))
    return variable


def create_prefix_dict(prefix_s: dict):
    prefix_dict = {}
    for index in range(1, rows_filled_out(prefix_s)):
        try:
            prefix_dict[read_cell(prefix_s, index, 0)] = read_cell(prefix_s, index, 1)
        except Exception as e:
            f"Prefix Processing Error, at the prefix sheet, row: {index}, {e}"
    return prefix_dict


def extract_concept_scheme(concept_scheme_sheet: dict, prefix: dict):
    cs = models.ConceptScheme(
        uri=using_prefix_non_list_output(read_cell(concept_scheme_sheet, 0), prefix),
        title=read_cell(concept_scheme_sheet, 1),
        description=read_cell(concept_scheme_sheet, 2),
        created=read_cell(concept_scheme_sheet, 3),
        modified=read_cell(concept_scheme_sheet, 4),
        creator=read_cell(concept_scheme_sheet, 5),
        publisher=read_cell(concept_scheme_sheet, 6),
        version=read_cell(concept_scheme_sheet, 7),
        provenance=read_cell(concept_scheme_sheet, 8),
        custodian=read_cell(concept_scheme_sheet, 9),
        pid=read_cell(concept_scheme_sheet, 10),
    )
    return cs


def extract_concepts(
    concepts_sheet: dict, additional_concepts_sheet: dict, prefix: dict
):
    concepts = []
    for index in range(2, rows_filled_out(concepts_sheet)):
        try:
            c = models.Concept(
                uri=using_prefix_non_list_output(
                    read_cell(concepts_sheet, index, 0), prefix
                ),
                pref_label=read_cell(concepts_sheet, index, 1),
                pl_language_code=split_and_tidy_to_strings(
                    read_cell(concepts_sheet, index, 2)
                ),
                definition=read_cell(concepts_sheet, index, 3),
                def_language_code=split_and_tidy_to_strings(
                    read_cell(concepts_sheet, index, 4)
                ),
                alt_labels=split_and_tidy_to_strings(
                    read_cell(concepts_sheet, index, 5)
                ),
                children=using_prefix_list_output(
                    split_and_tidy_to_strings(read_cell(concepts_sheet, index, 6)),
                    prefix,
                ),
                provenance=read_cell(concepts_sheet, index, 7),
                home_vocab_uri=using_prefix_non_list_output(
                    read_cell(concepts_sheet, index, 8), prefix
                ),
                # additional concepts sheet page
                related_match=using_prefix_list_output(
                    split_and_tidy_to_strings(
                        read_cell(additional_concepts_sheet, index, 0)
                    ),
                    prefix,
                ),
                close_match=using_prefix_list_output(
                    split_and_tidy_to_strings(
                        read_cell(additional_concepts_sheet, index, 1)
                    ),
                    prefix,
                ),
                exact_match=using_prefix_list_output(
                    split_and_tidy_to_strings(
                        read_cell(additional_concepts_sheet, index, 2)
                    ),
                    prefix,
                ),
                narrow_match=using_prefix_list_output(
                    split_and_tidy_to_strings(
                        read_cell(additional_concepts_sheet, index, 3)
                    ),
                    prefix,
                ),
                broad_match=using_prefix_list_output(
                    split_and_tidy_to_strings(
                        read_cell(additional_concepts_sheet, index, 4)
                    ),
                    prefix,
                ),
            )
            concepts.append(c)
        except ValidationError as e:
            raise ConversionError(
                f" Concept Processing error likely at sheet 'Concepts*' or 'Additional Concept Features', "
                f"row {index + 1}, and has {e} "
            )
    return concepts


def extract_collections(collection_sheet: dict, prefix: dict):
    collections = []
    row = rows_filled_out(collection_sheet)
    for index in range(2, row):
        try:
            c = models.Collection(
                uri=using_prefix_non_list_output(
                    read_cell(collection_sheet, index, 0), prefix
                ),
                pref_label=read_cell(collection_sheet, index, 1),
                definition=read_cell(collection_sheet, index, 2),
                members=using_prefix_list_output(
                    split_and_tidy_to_strings(read_cell(collection_sheet, index, 3)),
                    prefix,
                ),
                provenance=read_cell(collection_sheet, index, 4),
            )
            collections.append(c)
        except ValidationError as e:
            raise ConversionError(
                f" Collection Processing Error likely at the 'Collections' sheet, row {row}, and has {e}"
            )
    return collections


def excel_to_rdf_gsheets(
    uri: str,
    key: str,
    file_to_convert_path: Path,
    output_type: Literal["file", "string", "graph"] = "file",
    output_file_path=None,
    output_format: Literal["turtle", "xml", "json-ld"] = "turtle",
    profile="vocpub",
    error_level=1,
    message_level=1,
    log_file=None,
    validate=False,
):
    # get spreadsheet id from uri
    global SPREADSHEET_ID
    SPREADSHEET_ID = get_spreadsheetid_from_URI(uri)

    global KEY
    KEY = key

    global client
    client = build("sheets", "v4", developerKey=KEY)

    global sheet
    sheet = client.spreadsheets()

    # Template Version Validation
    template_version = get_template_version_gsheets(SPREADSHEET_ID)
    if template_version not in ACCEPTED_TEMPLATE_VERSIONS:
        raise ValueError(
            f"Unknown Template Version. Known Template Versions for Google Sheets are "
            f"{', '.join(ACCEPTED_TEMPLATE_VERSIONS)}, however you supplied {template_version}"
        )

    # Template Version 0.4.3
    if template_version == "0.4.3":
        try:
            concept_scheme_sheet = (
                sheet.values()
                .get(spreadsheetId=SPREADSHEET_ID, range="Concept Scheme*!B2:B12")
                .execute()
            )
            concept_sheet = (
                sheet.values()
                .get(spreadsheetId=SPREADSHEET_ID, range="Concepts*!A:I")
                .execute()
            )
            additional_concept_sheet = (
                sheet.values()
                .get(
                    spreadsheetId=SPREADSHEET_ID,
                    range="Additional Concept Features!B:F",
                )
                .execute()
            )
            collection_sheet = (
                sheet.values()
                .get(spreadsheetId=SPREADSHEET_ID, range="Collections!A:E")
                .execute()
            )
            prefix_sheet = (
                sheet.values()
                .get(spreadsheetId=SPREADSHEET_ID, range="Prefix Sheet!A:B")
                .execute()
            )

            prefix = create_prefix_dict(prefix_sheet)
            concept_scheme = extract_concept_scheme(concept_scheme_sheet, prefix)
            concept = extract_concepts(concept_sheet, additional_concept_sheet, prefix)
            collections = extract_collections(collection_sheet, prefix)
        except Exception as e:
            print(e)

    # Build the total vocab
    vocab_graph = models.Vocabulary(
        concept_scheme=concept_scheme, concepts=concept, collections=collections
    ).to_graph()

    # validate with vocpub
    if validate:
        validate_with_profile(
            vocab_graph,
            profile=profile,
            error_level=error_level,
            message_level=message_level,
            log_file=log_file,
        )

    # Write out the file
    if output_type == "graph":
        return vocab_graph
    elif output_type == "string":
        return vocab_graph.serialize(format=output_format)
    else:  # output_format == "file":
        if output_file_path is not None:
            dest = output_file_path
        else:
            if output_format == "xml":
                suffix = ".rdf"
            elif output_format == "json-ld":
                suffix = ".json-ld"
            else:
                suffix = ".ttl"
            dest = file_to_convert_path.with_suffix(suffix)
        vocab_graph.serialize(destination=str(dest), format=output_format)
        return dest


if __name__ == "__main__":
    pass



import sys
from pathlib import Path

import pytest
from rdflib import URIRef
from rdflib.namespace import RDF, SKOS

sys.path.append(str(Path(__file__).parent.parent.absolute() / "vocexcel"))
from vocexcel.convert_070 import (
    extract_concept_scheme,
    extract_concepts,
    extract_prefixes,
)
from vocexcel.utils import load_workbook

TESTS_DATA_DIR_PATH = Path(__file__).parent.absolute() / "data"


@pytest.fixture
def get_excel():
    return load_workbook(TESTS_DATA_DIR_PATH / "070_long.xlsx")


def test_extract_prefixes(get_excel):
    wb = get_excel
    n = extract_prefixes(wb["Prefixes"])

    assert n == {"ex:": "http://example.com/"}


def test_extract_concept_scheme(get_excel):
    wb = get_excel
    prefixes = extract_prefixes(wb["Prefixes"])
    cs, cs_iri = extract_concept_scheme(wb["Concept Scheme"], prefixes)

    assert cs_iri == URIRef(
        "https://pid.geoscience.gov.au/def/voc/ga/GeomorphologySettings"
    )


def test_extract_concepts(get_excel):
    wb = get_excel
    prefixes = extract_prefixes(wb["Prefixes"])
    cs, cs_iri = extract_concept_scheme(wb["Concept Scheme"], prefixes)
    cons = extract_concepts(wb["Concepts"], prefixes, cs_iri)

    concepts = []
    for c in cons.subjects(RDF.type, SKOS.Concept):
        concepts.append(c)

    assert (
        URIRef(
            "https://pid.geoscience.gov.au/def/voc/ga/GeomorphologySettings/Mass_Movement_Process"
        )
        in concepts
    )

    assert len(concepts) == 374


# @pytest.mark.xfail(reason="Failing since 0.6.2.")
# def test_extract_collections():
#     wb = load_workbook(Path(__file__).parent / "060_simple.xlsx")
#     prefixes = extract_prefixes(wb["Prefixes"])
#     cols = extract_collections(wb["Collections"], prefixes)
#     # print(cols.serialize(format="longturtle"))
#     expected = """
#         PREFIX dcterms: <http://purl.org/dc/terms/>
#         PREFIX ex: <http://example.com/>
#         PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
#
#         ex:small-chickens
#             a skos:Concept ;
#             dcterms:provenance "Just made up" ;
#             skos:definition "Breeds of small chickens"@en ;
#             skos:member
#                 ex:bantam ,
#                 ex:silkie ;
#             skos:prefLabel "Small Chickens"@en ;
#         .
#         """
#     g2 = Graph().parse(data=expected)
#     # print((g2 - cols).serialize(format="longturtle"))
#     # print("++++++++++++++")
#     # print((cols - g2).serialize(format="longturtle"))
#     # print(g2.serialize(format="longturtle"))
#     assert compare.isomorphic(cols, g2)
#
#
# def test_extract_additions_concept_properties():
#     wb = load_workbook(Path(__file__).parent / "060_simple.xlsx")
#     prefixes = extract_prefixes(wb["Prefixes"])
#     extra = extract_additions_concept_properties(
#         wb["Additional Concept Properties"], prefixes
#     )
#     expected = """
#         PREFIX ex: <http://example.com/>
#         PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
#
#         ex:bantam
#             skos:notation "ban"^^ex:other-system, "B"^^<http://system-x.com> ;
#             skos:relatedMatch <http://other-voc.com/bantam> ;
#         .
#         """
#     g2 = Graph().parse(data=expected)
#     assert compare.isomorphic(extra, g2)
#
#
# @pytest.mark.xfail(reason="Failing since 0.6.2.")
# def test_rdf_to_excel():
#     wb = load_workbook(Path(__file__).parent / "060_simple.xlsx")
#     g = excel_to_rdf(wb, output_format="graph")
#     # print(g.serialize(format="longturtle"))
#     expected = """
#         PREFIX cdterms: <http://purl.org/dc/terms/>
#         PREFIX ch: <http://example.com/chickens/>
#         PREFIX dcat: <http://www.w3.org/ns/dcat#>
#         PREFIX ex: <http://example.com/>
#         PREFIX isoroles: <http://def.isotc211.org/iso19115/-1/2018/CitationAndResponsiblePartyInformation/code/CI_RoleCode/>
#         PREFIX owl: <http://www.w3.org/2002/07/owl#>
#         PREFIX prov: <http://www.w3.org/ns/prov#>
#         PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
#         PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
#         PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
#
#         ex:chickens
#             a skos:ConceptScheme ;
#             cdterms:created "2022-12-01"^^xsd:date ;
#             cdterms:creator [
#                     a prov:Agent ;
#                     rdfs:label "Nicholas Car"
#                 ] ;
#             cdterms:identifier "chickens"^^xsd:token ;
#             cdterms:modified "2022-12-01"^^xsd:date ;
#             cdterms:publisher <http://linked.data.gov.au/org/agldwg> ;
#             owl:versionIRI ch:0.0.1 ;
#             owl:versionInfo "0.0.1" ;
#             skos:definition "A vocabulary of breeds of chicken"@en ;
#             skos:prefLabel "Chickens Breeds"@en ;
#             prov:qualifiedAttribution [
#                     dcat:hadRole isoroles:custodian ;
#                     prov:agent [
#                             a prov:Agent ;
#                             rdfs:label "Nicholas Car"
#                         ]
#                 ] ;
#         .
#
#         ex:rhode-island-red
#             a skos:Concept ;
#             cdterms:provenance "Taken from Wikipedia" ;
#             cdterms:source "https://en.wikipedia.org/wiki/Rhode_Island_Red"^^xsd:anyURI ;
#             skos:definition "The Rhode Island Red is an American breed of domestic chicken. It is the state bird of Rhode Island."@en ;
#             skos:prefLabel "Rhode Island Red"@en ;
#         .
#
#         ex:small-chickens
#             a skos:Concept ;
#             cdterms:provenance "Just made up" ;
#             skos:definition "Breeds of small chickens"@en ;
#             skos:member
#                 ex:bantam ,
#                 ex:silkie ;
#             skos:prefLabel "Small Chickens"@en ;
#         .
#
#         ex:bantam
#             a skos:Concept ;
#             skos:notation "ban"^^ex:other-system, "B"^^<http://system-x.com> ;
#             cdterms:source "https://en.wikipedia.org/wiki/Bantam_(poultry)"^^xsd:anyURI ;
#             skos:definition "A bantam is any small variety of fowl, usually of chicken or duck."@en ;
#             skos:narrower ex:silkie ;
#             skos:prefLabel "Bantam"@en ;
#             skos:relatedMatch <http://other-voc.com/bantam> ;
#         .
#
#         <http://linked.data.gov.au/org/agldwg>
#             a prov:Agent ;
#             rdfs:label "Agldwg" ;
#         .
#
#         ex:silkie
#             a skos:Concept ;
#             skos:altLabel
#                 "Chinese silk chicken"@en ,
#                 "Silky"@en ;
#             skos:definition "The Silkie (also known as the Silky or Chinese silk chicken) is a breed of chicken named for its atypically fluffy plumage, which is said to feel like silk and satin."@en ;
#             skos:prefLabel "Silkie"@en ;
#         .
#     """
#     g2 = Graph().parse(data=expected)
#     assert compare.isomorphic(g, g2)



import sys
from pathlib import Path

from rdflib import URIRef
from rdflib.namespace import SDO, RDF, SKOS

sys.path.append(str(Path(__file__).parent.parent.absolute() / "vocexcel"))
from vocexcel import convert

TEMPLATES_DIR_PATH = Path(__file__).parent.parent.absolute() / "vocexcel/templates"
TESTS_DATA_DIR_PATH = Path(__file__).parent.absolute() / "data"


def test_0812():
    g = convert.excel_to_rdf(TESTS_DATA_DIR_PATH / "0812.xlsx", output_format="graph")

    assert (
        URIRef("http://example.com/voc/rock-types"),
        SDO.status,
        URIRef("https://linked.data.gov.au/def/reg-statuses/experimental"),
    ) in g

    assert (
        URIRef("http://example.com/voc/rock-types/igenous"),
        RDF.type,
        SKOS.Concept,
    ) in g

    assert (
        URIRef("http://example.com/voc/rock-types/metamorphic"),
        RDF.type,
        SKOS.Concept,
    ) in g

    assert (
        URIRef(
            "https://linked.data.gov.au/def/gswa-rock-classification-scheme/sedimentary"
        ),
        RDF.type,
        SKOS.Concept,
    ) in g

    assert (
        URIRef("http://example.com/voc/rock-types/marble"),
        RDF.type,
        SKOS.Concept,
    ) in g

    print(g.serialize(format="longturtle"))


def test_0812_ntgs():
    g = convert.excel_to_rdf(
        TESTS_DATA_DIR_PATH / "0812_ntgs.xlsx", output_format="graph"
    )

    assert (
        URIRef("https://linked.data.gov.au/def/cox-classification/carbonatite"),
        SKOS.topConceptOf,
        URIRef("https://linked.data.gov.au/def/cox-classification"),
    ) in g

    assert (
        URIRef("https://linked.data.gov.au/def/cox-classification/carbonatite"),
        SKOS.narrower,
        URIRef("https://linked.data.gov.au/def/cox-classification/carbonate-hosted-zn"),
    ) in g

    assert (
        URIRef("https://linked.data.gov.au/def/cox-classification/carbonate-hosted-zn"),
        SKOS.broader,
        URIRef("https://linked.data.gov.au/def/cox-classification/carbonatite"),
    ) in g

    assert (
        URIRef("https://linked.data.gov.au/def/cox-classification/carbonate-hosted-zn"),
        SKOS.inScheme,
        URIRef("https://linked.data.gov.au/def/cox-classification"),
    ) in g

    # print(g.serialize(format="longturtle"))
